import collections
import json
import re
import traceback
from collections import defaultdict
from io import StringIO

import numpy as np
import pandas as pd
from loguru import logger

from .common import load_obj, save_obj
from .constants import COMPOUND_DATABASE_CHEBI, EXTERNAL_KEGG_TO_CHEBI, GENE_PK, PROTEIN_PK, REACTION_PK, \
    COMPOUND_PK, PATHWAY_PK, COMPOUND_DATABASE_KEGG, NA, EXTERNAL_GENE_NAMES, EXTERNAL_COMPOUND_NAMES, GENOMICS, \
    PROTEOMICS, METABOLOMICS, REACTIONS, PATHWAYS, GENES_TO_PROTEINS, PROTEINS_TO_REACTIONS, COMPOUNDS_TO_REACTIONS, \
    REACTIONS_TO_PATHWAYS, IDENTIFIER_COL, PIMP_PEAK_ID_COL, GROUP_COL, DEFAULT_GROUP_NAME, PADJ_COL_PREFIX, \
    CHEBI_BFS_RELATION_DICT, \
    FC_COL_PREFIX, SAMPLE_COL, DATA_DIR, CHEBI_RELATION_TSV
from .metadata import get_gene_names, get_compound_metadata, clean_label
from .reactome import ensembl_to_uniprot, uniprot_to_reaction, compound_to_reaction, \
    reaction_to_pathway, reaction_to_uniprot, reaction_to_compound, uniprot_to_ensembl

Relation = collections.namedtuple('Relation', 'keys values mapping_list')


# from pyMultiOmics.constants import IDENTIFIER_COL

def prepare_input(data_df):
    if data_df is None:
        return None

    # remove index, since all our old codes below assume this
    data_df = data_df.reset_index()

    # always set the first column to be the Identifier column
    rename = {data_df.columns.values[0]: IDENTIFIER_COL}
    data_df = data_df.rename(columns=rename)

    # assume id is in the first column and is a string
    data_df.iloc[:, 0] = data_df.iloc[:, 0].astype(str)
    return data_df


def reactome_mapping(observed_gene_df, observed_protein_df, observed_compound_df,
                     compound_database_str, species_list, metabolic_pathway_only,
                     include_related_chebi):
    observed_gene_df = prepare_input(observed_gene_df)
    observed_protein_df = prepare_input(observed_protein_df)
    observed_compound_df = prepare_input(observed_compound_df)

    ### all the ids that we have from the user ###
    observed_gene_ids = get_ids_from_dataframe(observed_gene_df)
    observed_protein_ids = get_ids_from_dataframe(observed_protein_df)
    observed_compound_ids = get_ids_from_dataframe(observed_compound_df)

    # try to convert all kegg ids to chebi ids, if possible
    KEGG_2_CHEBI = load_obj(EXTERNAL_KEGG_TO_CHEBI)
    not_found = []
    for cid in observed_compound_ids:
        if cid not in KEGG_2_CHEBI:
            not_found.append(cid)
            KEGG_2_CHEBI[cid] = cid

    if len(not_found) > 0:
        logger.debug('Not found: %d when converting kegg -> chebi ids' % len(not_found))
    if observed_compound_df is not None:
        if compound_database_str == COMPOUND_DATABASE_CHEBI:
            # if we have a df with KEGG ids, try to map this to chebi ids
            observed_compound_df.iloc[:, 0] = observed_compound_df.iloc[:, 0].map(
                KEGG_2_CHEBI)  # assume 1st column is id
            # get related chebi ids if necessary so we get more hits when mapping compounds
            if include_related_chebi:
                logger.info('Including related chebi ids')
                observed_compound_df = get_related_chebi(observed_compound_df)
        observed_compound_df = remove_dupes(observed_compound_df)
        observed_compound_ids = get_ids_from_dataframe(observed_compound_df)

        logger.info('There are %d observed compound ids' % len(observed_compound_ids))

    ### map genes -> proteins ###
    logger.info('Mapping genes -> proteins')
    gene_2_proteins_mapping, _ = ensembl_to_uniprot(observed_gene_ids, species_list)
    gene_2_proteins = make_relations(gene_2_proteins_mapping, GENE_PK, PROTEIN_PK, value_key=None)

    ### maps proteins -> reactions ###
    logger.info('Mapping proteins -> reactions')
    protein_ids_from_genes = gene_2_proteins.values
    known_protein_ids = list(set(observed_protein_ids + protein_ids_from_genes))
    protein_2_reactions_mapping, _ = uniprot_to_reaction(known_protein_ids, species_list)
    protein_2_reactions = make_relations(protein_2_reactions_mapping, PROTEIN_PK, REACTION_PK,
                                         value_key='reaction_id')

    ### maps compounds -> reactions ###
    logger.info('Mapping compounds -> reactions')
    compound_2_reactions_mapping, _ = compound_to_reaction(observed_compound_ids, species_list)
    compound_2_reactions = make_relations(compound_2_reactions_mapping, COMPOUND_PK, REACTION_PK,
                                          value_key='reaction_id')

    ### maps reactions -> metabolite pathways ###
    logger.info('Mapping reactions -> pathways')
    reaction_ids_from_proteins = protein_2_reactions.values
    reaction_ids_from_compounds = compound_2_reactions.values
    reaction_ids = list(set(reaction_ids_from_proteins + reaction_ids_from_compounds))
    reaction_2_pathways_mapping, reaction_2_pathways_id_to_names = reaction_to_pathway(reaction_ids,
                                                                                       species_list,
                                                                                       metabolic_pathway_only)
    reaction_2_pathways = make_relations(reaction_2_pathways_mapping, REACTION_PK, PATHWAY_PK,
                                         value_key='pathway_id')

    ### maps reactions -> proteins ###
    logger.info('Mapping reactions -> proteins')
    mapping, _ = reaction_to_uniprot(reaction_ids, species_list)
    reaction_2_proteins = make_relations(mapping, REACTION_PK, PROTEIN_PK, value_key=None)
    protein_2_reactions = merge_relation(protein_2_reactions, reverse_relation(reaction_2_proteins))
    all_protein_ids = protein_2_reactions.keys

    ### maps reactions -> compounds ###
    logger.info('Mapping reactions -> compounds')
    if compound_database_str == COMPOUND_DATABASE_KEGG:
        use_kegg = True
    else:
        use_kegg = False
    reaction_2_compounds_mapping, reaction_to_compound_id_to_names = reaction_to_compound(reaction_ids, species_list,
                                                                                          use_kegg)
    reaction_2_compounds = make_relations(reaction_2_compounds_mapping, REACTION_PK, COMPOUND_PK, value_key=None)
    compound_2_reactions = merge_relation(compound_2_reactions, reverse_relation(reaction_2_compounds))
    all_compound_ids = compound_2_reactions.keys

    ### map proteins -> genes ###
    logger.info('Mapping proteins -> genes')
    mapping, _ = uniprot_to_ensembl(all_protein_ids, species_list)
    protein_2_genes = make_relations(mapping, PROTEIN_PK, GENE_PK, value_key=None)
    gene_2_proteins = merge_relation(gene_2_proteins, reverse_relation(protein_2_genes))
    all_gene_ids = gene_2_proteins.keys

    ### add links ###

    # map NA to NA
    gene_2_proteins = add_links(gene_2_proteins, GENE_PK, PROTEIN_PK, [NA], [NA])
    protein_2_reactions = add_links(protein_2_reactions, PROTEIN_PK, REACTION_PK, [NA], [NA])
    compound_2_reactions = add_links(compound_2_reactions, COMPOUND_PK, REACTION_PK, [NA], [NA])
    reaction_2_pathways = add_links(reaction_2_pathways, REACTION_PK, PATHWAY_PK, [NA], [NA])

    # map genes that have no proteins to NA
    gene_pk_list = [x for x in all_gene_ids if x not in gene_2_proteins.keys]
    gene_2_proteins = add_links(gene_2_proteins, GENE_PK, PROTEIN_PK, gene_pk_list, [NA])

    # map proteins that have no genes to NA
    protein_pk_list = [x for x in all_protein_ids if x not in gene_2_proteins.values]
    gene_2_proteins = add_links(gene_2_proteins, GENE_PK, PROTEIN_PK, [NA], protein_pk_list)

    # map proteins that have no reactions to NA
    protein_pk_list = [x for x in all_protein_ids if x not in protein_2_reactions.keys]
    protein_2_reactions = add_links(protein_2_reactions, PROTEIN_PK, REACTION_PK, protein_pk_list, [NA])

    # map reactions that have no proteins to NA
    reaction_pk_list = [x for x in reaction_ids if x not in protein_2_reactions.values]
    protein_2_reactions = add_links(protein_2_reactions, PROTEIN_PK, REACTION_PK, [NA], reaction_pk_list)

    # map compounds that have no reactions to NA
    compound_pk_list = [x for x in all_compound_ids if x not in compound_2_reactions.keys]
    compound_2_reactions = add_links(compound_2_reactions, COMPOUND_PK, REACTION_PK, compound_pk_list, [NA])

    # map reactions that have no compounds to NA
    reaction_pk_list = [x for x in reaction_ids if x not in compound_2_reactions.values]
    compound_2_reactions = add_links(compound_2_reactions, COMPOUND_PK, REACTION_PK, [NA], reaction_pk_list)

    # map reactions that have no pathways to NA
    reaction_pk_list = [x for x in reaction_ids if x not in reaction_2_pathways.keys]
    reaction_2_pathways = add_links(reaction_2_pathways, REACTION_PK, PATHWAY_PK, reaction_pk_list, [NA])

    GTF_DICT = load_obj(EXTERNAL_GENE_NAMES)
    metadata_map = get_gene_names(all_gene_ids, GTF_DICT)
    genes_json = pk_to_json(GENE_PK, 'gene_id', all_gene_ids, metadata_map, observed_gene_df,
                            observed_ids=observed_gene_ids)
    gene_2_proteins_json = json.dumps(gene_2_proteins.mapping_list)

    # metadata_map = get_uniprot_metadata_online(uniprot_ids)
    proteins_json = pk_to_json('protein_pk', 'protein_id', all_protein_ids, metadata_map, observed_protein_df,
                               observed_ids=observed_protein_ids)
    protein_2_reactions_json = json.dumps(protein_2_reactions.mapping_list)

    # TODO: this feels like a very bad way to implement this
    # We need to deal with uploaded peak data from PiMP, which contains a lot of duplicate identifications per peak
    KEGG_ID_2_DISPLAY_NAMES = load_obj(EXTERNAL_COMPOUND_NAMES)
    metadata_map = get_compound_metadata(all_compound_ids, KEGG_ID_2_DISPLAY_NAMES, reaction_to_compound_id_to_names)
    try:
        mapping = get_mapping(observed_compound_df)
    except KeyError:
        mapping = None
    except AttributeError:
        mapping = None

    compounds_json = pk_to_json('compound_pk', 'compound_id', all_compound_ids, metadata_map, observed_compound_df,
                                observed_ids=observed_compound_ids, mapping=mapping)
    if mapping:
        compound_2_reactions = expand_relation(compound_2_reactions, mapping, 'compound_pk')
    compound_2_reactions_json = json.dumps(compound_2_reactions.mapping_list)

    metadata_map = {}
    for name in reaction_2_pathways_id_to_names:
        tok = reaction_2_pathways_id_to_names[name]['name']
        filtered = clean_label(tok)
        species = reaction_2_pathways_id_to_names[name]['species']
        metadata_map[name] = {'display_name': filtered, 'species': species}

    reaction_count_df = None
    pathway_count_df = None

    pathway_ids = reaction_2_pathways.values
    reactions_json = pk_to_json('reaction_pk', 'reaction_id', reaction_ids, metadata_map, reaction_count_df,
                                has_species=True)
    pathways_json = pk_to_json('pathway_pk', 'pathway_id', pathway_ids, metadata_map, pathway_count_df,
                               has_species=True)
    reaction_2_pathways_json = json.dumps(reaction_2_pathways.mapping_list)

    results = {
        GENOMICS: genes_json,
        PROTEOMICS: proteins_json,
        METABOLOMICS: compounds_json,
        REACTIONS: reactions_json,
        PATHWAYS: pathways_json,
        GENES_TO_PROTEINS: gene_2_proteins_json,
        PROTEINS_TO_REACTIONS: protein_2_reactions_json,
        COMPOUNDS_TO_REACTIONS: compound_2_reactions_json,
        REACTIONS_TO_PATHWAYS: reaction_2_pathways_json,
    }
    return results


def get_mapping(observed_compound_df):
    mapping = defaultdict(list)
    for idx, row in observed_compound_df.iterrows():
        identifier = row[IDENTIFIER_COL]
        peak_id = row[PIMP_PEAK_ID_COL]
        mapping[identifier].append('%s_%s' % (identifier, peak_id))
    return dict(mapping)


def csv_to_dataframe(csv_str):
    # extract group, if any
    filtered_str = ''
    group_str = None
    for line in csv_str.splitlines():  # go through all lines and remove the line containing the grouping info
        if re.match(GROUP_COL, line, re.I):
            group_str = line
        else:
            filtered_str += line + '\n'

    # extract id values
    data = StringIO(filtered_str)
    try:
        data_df = pd.read_csv(data)
        data_df.columns = data_df.columns.str.replace('.',
                                                      '_')  # replace period with underscore to prevent alasql breaking
        data_df.columns = data_df.columns.str.replace('-',
                                                      '_')  # replace dash with underscore to prevent alasql breaking
        data_df.columns = data_df.columns.str.replace('#', '')  # remove funny characters
        rename = {data_df.columns.values[0]: IDENTIFIER_COL}
        for i in range(len(data_df.columns.values[1:])):  # sql doesn't like column names starting with a number
            col_name = data_df.columns.values[i]
            if col_name[0].isdigit():
                new_col_name = '_' + col_name  # append an underscore in front of the column name
                rename[col_name] = new_col_name
        data_df = data_df.rename(columns=rename)
        data_df.iloc[:, 0] = data_df.iloc[:, 0].astype(str)  # assume id is in the first column and is a string
        id_list = data_df.iloc[:, 0].values.tolist()
    except pd.errors.EmptyDataError:
        data_df = None
        id_list = []

    # create grouping dataframe
    group_df = None
    if data_df is not None:
        sample_data = data_df.columns.values
        if group_str is not None:
            group_data = group_str.split(',')
        else:
            num_samples = len(sample_data)
            group_data = [DEFAULT_GROUP_NAME for x in
                          range(num_samples)]  # assigns a default group if nothing specified

        # skip non-measurement columns
        filtered_sample_data = []
        filtered_group_data = []
        for i in range(len(sample_data)):
            sample_name = sample_data[i]
            if sample_name == IDENTIFIER_COL or \
                    sample_name == PIMP_PEAK_ID_COL or \
                    sample_name.startswith(PADJ_COL_PREFIX) or \
                    sample_name.startswith(FC_COL_PREFIX):
                continue
            filtered_sample_data.append(sample_data[i])
            filtered_group_data.append(group_data[i])

        # convert to dataframe
        if len(filtered_group_data) > 0:
            group_df = pd.DataFrame(list(zip(filtered_sample_data, filtered_group_data)),
                                    columns=[SAMPLE_COL, GROUP_COL])

    return data_df, group_df, id_list


def get_ids_from_dataframe(df):
    if df is None:
        return []
    else:
        return df.iloc[:, 0].values.tolist()  # id is always the first column


def get_related_chebi(cmpd_df):
    """
    A method to add rows for related Chebi_ids given an Identifier that has related chebi_ids.
    The rows Chebi_id: related_chebi_id are duplicated in terms of the intenisty data
    :param cmpd_df: A DF with Chebi_ids as the Identifiers.
    :return: cmpd_df including the rows of intensity data for the related chebis.
    """
    cmpd_df = cmpd_df.copy()

    # ensure index type is set to string, since get_chebi_relation_dict also returns string as the keys
    cmpd_df.index = cmpd_df.index.map(str)
    # cmpd_df = cmpd_df.reset_index()
    original_cmpds = set(cmpd_df[IDENTIFIER_COL])  # used for checking later

    # construct the related chebi dict
    chebi_rel_dict = get_chebi_relation_dict()

    # loop through each row in cmpd_data
    with_related_data = []
    for ix, row in cmpd_df.iterrows():

        # add the current row we're looping
        current_identifier = row[IDENTIFIER_COL]
        with_related_data.append(row)

        # check if there are related compounds to add
        if current_identifier in chebi_rel_dict:
            # if yes, get the related compounds
            chebi_list = chebi_rel_dict[current_identifier]
            for c in chebi_list:
                # add the related chebi, but only if it's not already present in the original compound
                if c not in original_cmpds:
                    current_row = row.copy()
                    current_row[IDENTIFIER_COL] = c
                    with_related_data.append(current_row)

    # combine all the rows into a single dataframe
    df = pd.concat(with_related_data, axis=1).T
    # df = df.set_index(IDENTIFIER_COL)
    logger.info('Inserted %d related compounds' % (len(df) - len(cmpd_df)))

    # Remove any rows with duplicate Chebi_ids
    # rm_dups_df = remove_dupes(df)

    return df


def remove_dupes(df):
    """
    :param df: A dataframe with an Identifier column and intensity data.
    :return: The same dataframe with one row per identifier - with the row chosen of max sum of intensities
    across the row. Duplicate rows deleted.
    """
    # df = df.reset_index()

    # group df by the 'Identifier' column
    to_delete = []
    grouped = df.groupby(df[IDENTIFIER_COL])
    for identifier, group_df in grouped:

        # if there are multiple rows sharing the same identifier
        if len(group_df) > 1:
            # remove 'Identifier' column from the grouped df since it can't be summed
            group_df = group_df.drop(IDENTIFIER_COL, axis=1)

            # find the row with the largest sum across the row in the group
            idxmax = group_df.sum(axis=1).idxmax()

            # mark all the rows in the group for deletion, except the one with the largest sum
            temp = group_df.index.tolist()
            temp.remove(idxmax)
            to_delete.extend(temp)

    # actually do the deletion here
    logger.info('Removing %d rows with duplicate identifiers' % (len(to_delete)))
    df = df.drop(to_delete)
    # df = df.set_index(IDENTIFIER_COL)
    return df


def get_chebi_relation_dict():
    """
    A method to parse the chebi relation tsv and store the relationship we want in a dictionary
    :return: Dict with structure Chebi_id: [related_chebi_ids]
    """

    logger.info("Getting %s " % CHEBI_BFS_RELATION_DICT)

    chebi_bfs_relation_dict = load_obj(CHEBI_BFS_RELATION_DICT)

    if chebi_bfs_relation_dict is None:
        logger.info("Constructing %s " % CHEBI_BFS_RELATION_DICT)
        try:
            chebi_relation_df = pd.read_csv(CHEBI_RELATION_TSV, delimiter="\t")
        except FileNotFoundError as e:
            logger.error("data/relation.tsv must be present")
            raise e

        # List of relationship we want in the dictionary
        select_list = ["is_conjugate_base_of", "is_conjugate_acid_of", "is_tautomer_of"]
        chebi_select_df = chebi_relation_df[chebi_relation_df.TYPE.isin(select_list)]

        chebi_relation_dict = {}
        # Gather all the INIT_IDs into a dictionary so that each INIT_ID is unique
        for ix, row in chebi_select_df.iterrows():
            init_id = str(row.INIT_ID)
            final_id = str(row.FINAL_ID)
            if init_id in chebi_relation_dict.keys():
                # Append the final_id onto the existing values
                id_1 = chebi_relation_dict[init_id]
                joined_string = ", ".join([id_1, final_id])
                chebi_relation_dict[init_id] = joined_string
            else:  # make a new key entry for the dict
                chebi_relation_dict[init_id] = final_id

        # Change string values to a list.
        graph = {k: v.replace(" ", "").split(",") for k, v in chebi_relation_dict.items()}

        chebi_bfs_relation_dict = {}
        for k, v in graph.items():
            r_chebis = bfs_get_related(graph, k)
            r_chebis.remove(k)  # remove original key from list
            chebi_bfs_relation_dict[k] = r_chebis
        try:
            logger.info("saving chebi_relation_dict")
            save_obj(chebi_bfs_relation_dict, CHEBI_BFS_RELATION_DICT)

        except Exception as e:
            logger.error("Pickle didn't work because of %s " % e)
            traceback.print_exc()
            pass

    return chebi_bfs_relation_dict


def bfs_get_related(graph_dict, node):
    """
    :param graph: Dictionary of key: ['value'] pairs
    :param node: the key for which all related values should be returned
    :return: All related keys as a list
    """
    visited = []  # List to keep track of visited nodes.
    queue = []  # Initialize a queue
    related_keys = []

    visited.append(node)
    queue.append(node)

    while queue:
        k = queue.pop(0)
        related_keys.append(k)

        for neighbour in graph_dict[k]:
            if neighbour not in visited:
                visited.append(neighbour)
                queue.append(neighbour)

    return related_keys


def merge_relation(r1, r2):
    unique_keys = list(set(r1.keys + r2.keys))
    unique_values = list(set(r1.values + r2.values))
    mapping_list = r1.mapping_list + r2.mapping_list
    mapping_list = list(map(dict, set(map(lambda x: frozenset(x.items()), mapping_list))))  # removes duplicates, if any
    return Relation(keys=list(unique_keys), values=list(unique_values),
                    mapping_list=mapping_list)


def reverse_relation(rel):
    return Relation(keys=rel.values, values=rel.keys, mapping_list=rel.mapping_list)


def expand_relation(rel, mapping, pk_col):
    expanded_keys = substitute(rel.keys, mapping)
    expanded_values = substitute(rel.values, mapping)
    expanded_mapping_list = []
    for row in rel.mapping_list:
        expanded = expand_each(row, mapping, pk_col)
        if len(expanded) == 0:
            expanded = [row]
        expanded_mapping_list.extend(expanded)
    return Relation(keys=expanded_keys, values=expanded_values, mapping_list=expanded_mapping_list)


def substitute(my_list, mapping):
    new_list = []
    for x in my_list:
        if x in mapping:
            new_list.extend(mapping[x])
        else:
            new_list.append(x)
    return new_list


def expand_each(row, mapping, pk_col):
    results = []
    pk = row[pk_col]
    try:
        replacements = mapping[pk]
        for rep in replacements:
            new_row = without_keys(row, [pk_col])
            new_row[pk_col] = rep
            results.append(new_row)
    except KeyError:
        pass
    return results


# https://stackoverflow.com/questions/31433989/return-copy-of-dictionary-excluding-specified-keys
def without_keys(d, keys):
    return {x: d[x] for x in d if x not in keys}


def pk_to_json(pk_label, display_label, data, metadata_map, observed_df, has_species=False,
               observed_ids=None, mapping=None):
    if observed_df is not None:
        if PIMP_PEAK_ID_COL in observed_df.columns:  # if peak id is present, rename the identifier column to include it
            observed_df[IDENTIFIER_COL] = observed_df[IDENTIFIER_COL] + '_' + observed_df[PIMP_PEAK_ID_COL].astype(str)
        if mapping is not None:
            data = expand_data(data, mapping)

        observed_df = observed_df.set_index(IDENTIFIER_COL)  # set identifier as index
        observed_df = observed_df[~observed_df.index.duplicated(keep='first')]  # remove row with duplicate indices
        observed_df = observed_df.fillna(value=0)  # replace all NaNs with 0s

    output = []
    for item in sorted(data):

        if item == NA:
            continue  # handled below after this loop

        if '_' in item:
            tokens = item.split('_')
            assert len(tokens) == 2
            item = tokens[0]
            peak_id = tokens[1]
        else:
            peak_id = None

        # add observed status and the primary key label to row data
        row = {}
        if observed_ids is not None:
            if item in observed_ids:
                row['obs'] = True
            else:
                row['obs'] = False
        else:
            row['obs'] = None

        if peak_id:
            key = '%s_%s' % (item, peak_id)
            row[pk_label] = key
        else:
            row[pk_label] = item

        # add display label to row_data
        species = None
        if len(metadata_map) > 0 and item in metadata_map and metadata_map[item] is not None:
            if peak_id:
                label = '%s (%s)' % (metadata_map[item]['display_name'].capitalize(), peak_id)
            else:
                label = metadata_map[item]['display_name'].capitalize()
            # get the species if it's there too
            if has_species and 'species' in metadata_map[item]:
                species = metadata_map[item]['species']
        else:
            label = item  # otherwise use the item id as the label
        row[display_label] = label

        # add the remaining data columns to row
        if observed_df is not None:
            try:
                if peak_id:
                    observed_values = observed_df.loc[key].to_dict()
                else:
                    observed_values = observed_df.loc[item].to_dict()
            except KeyError:  # missing data
                observed_values = {}
                for col in observed_df.columns:
                    observed_values.update({col: None})
            observed_values.pop(PIMP_PEAK_ID_COL, None)  # remove pimp peakid column
            # convert numpy bool to python bool, else json serialisation will break
            for k, v in observed_values.items():
                if type(v) == np.bool_:
                    observed_values[k] = bool(v)
            row.update(observed_values)

        if has_species:
            row['species'] = species

        if row not in output:
            output.append(row)

    # add dummy entry
    row = {'obs': NA, pk_label: NA, display_label: NA}
    if has_species:
        row['species'] = NA

    if observed_df is not None:  # also add the remaining columns
        for col in observed_df.columns:
            if col == PIMP_PEAK_ID_COL:
                continue
            row.update({col: 0})

    if row not in output:
        output.append(row)

    output_json = json.dumps(output)
    return output_json


def expand_data(data, mapping):
    new_data = []
    for x in data:
        if x in mapping:
            new_data.extend(mapping[x])
        else:
            new_data.append(x)
    data = new_data
    return data


def make_relations(mapping, source_pk, target_pk, value_key=None):
    id_values = []
    mapping_list = []

    for key in mapping:
        value_list = mapping[key]

        # value_list can be either a list of strings or dictionaries
        # check if the first element is a dict, else assume it's a string
        assert len(value_list) > 0
        is_string = True
        first = value_list[0]
        if isinstance(first, dict):
            is_string = False

        # process each element in value_list
        for value in value_list:
            if is_string:  # value_list is a list of string
                actual_value = value
            else:  # value_list is a list of dicts
                assert value_key is not None, 'value_key is missing'
                actual_value = value[value_key]
            id_values.append(actual_value)
            row = {source_pk: key, target_pk: actual_value}
            mapping_list.append(row)

    unique_keys = set(mapping.keys())
    unique_values = set(id_values)

    return Relation(keys=list(unique_keys), values=list(unique_values),
                    mapping_list=mapping_list)


def add_dummy(relation, source_ids, target_ids, source_pk_label, target_pk_label):
    to_add = [x for x in source_ids if x not in relation.keys]
    relation = add_links(relation, source_pk_label, target_pk_label, to_add, [NA])

    # to_add = [x for x in target_ids if x not in relation.values]
    # relation = add_links(relation, source_pk_label, target_pk_label, [NA], to_add)

    # relation = add_links(relation, source_pk_label, target_pk_label, [NA], [NA])
    return relation


def add_links(relation, source_pk_label, target_pk_label, source_pk_list, target_pk_list):
    rel_mapping_list = list(relation.mapping_list)
    rel_keys = relation.keys
    rel_values = relation.values

    for s1 in source_pk_list:
        if s1 not in rel_keys: rel_keys.append(s1)
        for s2 in target_pk_list:
            rel_mapping_list.append({source_pk_label: s1, target_pk_label: s2})
            if s2 not in rel_keys: rel_values.append(s2)

    return Relation(keys=rel_keys, values=rel_values, mapping_list=rel_mapping_list)



import collections
import json
import urllib

from .common import truncate
from .constants import GENES, PROTEINS, COMPOUNDS, REACTIONS, PATHWAYS
from .harmonizomeapi import Harmonizome, Entity
from .metadata import get_single_ensembl_metadata_online, get_single_uniprot_metadata_online, \
    get_single_compound_metadata_online
from .reactome import get_reactome_description, get_reaction_entities, pathway_to_reactions


def get_info(node_id, data_type):
    if data_type == 'genes':
        res = get_ensembl_gene_info(node_id)
    elif data_type == 'proteins':
        res = get_uniprot_protein_info(node_id)
    elif data_type == 'compounds':
        res = get_kegg_metabolite_info(node_id)
    elif data_type == 'reactions':
        res = get_reactome_reaction_info(node_id)
    elif data_type == 'pathways':
        res = get_reactome_pathway_info(node_id)
    return res


def get_ensembl_gene_info(ensembl_id):
    display_name = ''
    try:
        metadata = get_single_ensembl_metadata_online(ensembl_id)
        display_name = metadata['display_name'] if metadata is not None and 'display_name' in metadata else ''

        infos = []
        if metadata is not None:
            # selected = ['description', 'species', 'biotype', 'db_type', 'logic_name', 'strand', 'start', 'end']
            selected = ['description', 'species']
            for key in selected:
                if key in metadata:
                    value = metadata[key]
                    if key == 'description':
                        value = value[0:value.index('[')]  # remove e.g. '[xxx]' from 'abhydrolase [xxx]'
                    infos.append({'key': key.title(), 'value': value})
    except TypeError:  # https://github.com/joewandy/GraphOmics/issues/9
        pass
    except KeyError:  # https://github.com/joewandy/GraphOmics/issues/10
        pass

    data = Harmonizome.get(Entity.GENE, name=display_name)
    try:
        description = data['description']
        infos.append({'key': 'Description', 'value': description})
    except KeyError:
        pass

    links = [
        {
            'text': 'Link to Ensembl',
            'href': 'https://www.ensembl.org/id/' + ensembl_id
        },
        {
            'text': 'Link to GeneCard',
            'href': 'https://www.genecards.org/cgi-bin/carddisp.pl?gene=' + display_name
        }
    ]
    for x in metadata['Transcript']:
        text = 'Transcript: ' + x['display_name']
        href = 'https://www.ensembl.org/id/' + x['id']
        links.append({'text': text, 'href': href})

    images = []
    data = {
        'infos': infos,
        'links': links,
        'images': images
    }
    return data


def get_uniprot_protein_info(uniprot_id):
    infos = []
    try:
        metadata = get_single_uniprot_metadata_online(uniprot_id)
        try:
            fullname = [x.text for x in metadata.soup.select('protein > recommendedname > fullname')][0]
        except IndexError:
            fullname = uniprot_id

        shortName = None
        try:
            shortname = [x.text for x in metadata.soup.select('protein > recommendedname > shortname')][0]
        except IndexError:
            pass

        if shortName is not None:
            infos.append({'key': 'Name', 'value': "{} ({})".format(fullname, shortname)})
        else:
            infos.append({'key': 'Name', 'value': "{}".format(fullname)})

        try:
            ecnumber = [x.text for x in metadata.soup.select('protein > recommendedname > ecnumber')][0]
            infos.append({'key': 'EC Number', 'value': 'EC' + ecnumber})
        except IndexError:
            pass

        # get comments
        selected = ['function', 'catalytic activity', 'enzyme regulation', 'subunit', 'pathway', 'miscellaneous',
                    'domain']
        for child in metadata.soup.find_all('comment'):
            try:
                if child['type'] in selected:
                    key = child['type'].title()
                    if key == 'Function':  # quick-hack
                        key = 'Summary'
                    infos.append({'key': key, 'value': truncate(child.text)})
            except KeyError:
                continue

        # gene ontologies
        go = []
        for child in metadata.soup.find_all('dbreference'):
            try:
                if child['type'] == 'GO':
                    props = child.find_all('property')
                    for prop in props:
                        if prop['type'] == 'term':
                            go.append(prop['value'].split(':')[1])
            except KeyError:
                continue
        go_str = '; '.join(sorted(go))
        go_str = truncate(go_str)
        infos.append({'key': 'Gene_ontologies', 'value': go_str})

    except TypeError:
        pass

    images = []
    with urllib.request.urlopen('https://swissmodel.expasy.org/repository/uniprot/' + uniprot_id + '.json') as url:
        data = json.loads(url.read().decode())
        for struct in data['result']['structures']:
            pdb_link = struct['coordinates']
            images.append(pdb_link)

    links = [
        {
            'text': 'Link to UniProt',
            'href': 'http://www.uniprot.org/uniprot/' + uniprot_id
        },
        {
            'text': 'Link to SWISS-MODEL',
            'href': 'https://swissmodel.expasy.org/repository/uniprot/' + uniprot_id
        }
    ]

    data = {
        'infos': infos,
        'links': links,
        'images': images
    }
    return data


def get_kegg_metabolite_info(compound_id):
    if '_' in compound_id:
        tokens = compound_id.split('_')
        assert len(tokens) == 2
        compound_id = tokens[0]
        peak_id = tokens[1]
    else:
        peak_id = None

    metadata = get_single_compound_metadata_online(compound_id)

    if compound_id.upper().startswith('C'):  # assume it's kegg

        # TODO: no longer used??!!

        infos = []
        selected = ['FORMULA']
        if metadata is not None:
            for key in selected:
                value = metadata[key]
                infos.append({'key': key, 'value': str(value)})

        key = 'PiMP Peak ID'
        value = peak_id
        infos.append({'key': key, 'value': str(value)})

        images = ['http://www.kegg.jp/Fig/compound/' + compound_id + '.gif']
        links = [
            {
                'text': 'Link to KEGG COMPOUND database',
                'href': 'http://www.genome.jp/dbget-bin/www_bget?cpd:' + compound_id
            }
        ]

    else:  # assume it's ChEBI

        def get_attribute(metadata, attrname):
            try:
                attr_val = getattr(metadata, attrname)
            except AttributeError:
                attr_val = ''
            return attr_val

        images = [
            'http://www.ebi.ac.uk/chebi/displayImage.do?defaultImage=true&imageIndex=0&chebiId=' + compound_id]
        links = [
            {
                'text': 'Link to ChEBI database',
                'href': 'https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:' + compound_id
            }
        ]

        infos = []
        key = 'PiMP Peak ID'
        value = peak_id
        infos.append({'key': key, 'value': str(value)})

        try:
            for db_link in metadata.DatabaseLinks:
                if 'KEGG COMPOUND' in str(db_link.type):
                    kegg_id = db_link.data
                    infos.append({'key': 'KEGG ID', 'value': kegg_id})
                    links.append({
                        'text': 'Link to KEGG COMPOUND database',
                        'href': 'http://www.genome.jp/dbget-bin/www_bget?cpd:' + kegg_id
                    })
        except AttributeError:
            pass

        try:
            infos.append({'key': 'FORMULA', 'value': metadata.Formulae[0].data})
        except AttributeError:
            pass
        selected = [
            ('ChEBI ID', 'chebiId'),
            ('Definition', 'definition'),
            ('Monoisotopic Mass', 'monoisotopicMass'),
            ('SMILES', 'smiles'),
            ('Inchi', 'inchi'),
            ('InchiKey', 'inchikey'),
        ]
        for key, attribute in selected:
            value = get_attribute(metadata, attribute)
            if value is not None:
                infos.append({'key': key, 'value': value})

    data = {
        'infos': infos,
        'images': images,
        'links': links
    }
    return data


def get_reactome_reaction_info(reactome_id):
    infos = []

    # res = get_reactome_content_service(reactome_id)
    # summary_str = res['summation'][0]['text']
    summary_str, is_inferred, original_species, inferred_species = _get_summary_string(reactome_id)
    infos.append({'key': 'Summary', 'value': summary_str})

    infos.append({'key': 'Species', 'value': original_species})
    if is_inferred:
        inferred = 'Inferred from %s' % inferred_species
        infos.append({'key': 'Inferred', 'value': inferred})

    # get all the participants
    temp = collections.defaultdict(list)
    results = get_reaction_entities([reactome_id])[reactome_id]
    for res in results:
        entity_id = res[1]
        display_name = res[2]
        relationship_types = res[3]
        if len(relationship_types) == 1:  # ignore the sub-complexes
            rel = relationship_types[0]
            temp[rel].append((display_name, entity_id,))

    for k, v in temp.items():
        name_list, url_list = zip(
            *v)  # https://stackoverflow.com/questions/12974474/how-to-unzip-a-list-of-tuples-into-individual-lists
        url_list = map(lambda x: 'https://reactome.org/content/detail/' + x if x is not None else '', url_list)

        name_str = ';'.join(name_list)
        url_str = ';'.join(url_list)
        infos.append({'key': k.title(), 'value': name_str, 'url': url_str})

    images = [
        'https://reactome.org/ContentService/exporter/diagram/' + reactome_id + '.jpg?sel=' + reactome_id + "&quality=7"
    ]
    links = [
        {
            'text': 'Link to Reactome database',
            'href': 'https://reactome.org/content/detail/' + reactome_id
        }
    ]
    data = {
        'infos': infos,
        'images': images,
        'links': links,
    }
    return data


def get_reactome_pathway_info(pathway_id):
    mapping, id_to_names = pathway_to_reactions([pathway_id])
    pathway_reactions = mapping[pathway_id]

    reaction_list = map(lambda x: id_to_names[x], pathway_reactions)
    reaction_str = ';'.join(sorted(reaction_list))

    url_list = map(lambda x: 'https://reactome.org/content/detail/' + x, pathway_reactions)
    url_str = ';'.join(sorted(url_list))

    # res = get_reactome_content_service(pathway_id)
    # summary_str = res['summation'][0]['text']
    summary_str, is_inferred, original_species, inferred_species = _get_summary_string(pathway_id)

    infos = [{'key': 'Summary', 'value': summary_str}]
    infos.append({'key': 'Species', 'value': original_species})
    if is_inferred:
        inferred = 'Inferred from %s' % inferred_species
        infos.append({'key': 'Inferred', 'value': inferred})
    infos.append({'key': 'Reactions', 'value': reaction_str, 'url': url_str})

    images = [
        'https://reactome.org/ContentService/exporter/diagram/' + pathway_id + '.jpg?sel=' + pathway_id
    ]
    links = [
        {
            'text': 'Link to Reactome database',
            'href': 'https://reactome.org/content/detail/' + pathway_id
        },
        {
            'text': 'SBML Export',
            'href': 'https://reactome.org/ContentService/exporter/sbml/' + pathway_id + '.xml'
        }
    ]
    data = {
        'infos': infos,
        'images': images,
        'links': links,
    }
    return data


def _get_summary_string(reactome_id):
    desc, is_inferred = get_reactome_description(reactome_id, from_parent=False)
    original_species = desc[0]['species']
    if is_inferred:
        desc, _ = get_reactome_description(reactome_id, from_parent=True)
        inferred_species = desc[0]['species']
    else:
        inferred_species = original_species

    summary_list = []
    for s in desc:
        if s['summary_text'] is None:
            continue
        st = s['summary_text'].replace(';', ',')
        summary_list.append(truncate(st))
    summary_str = ';'.join(summary_list)
    return summary_str, is_inferred, original_species, inferred_species



"""Class for reading, parsing, and downloading data from the Harmonizome API.
"""

import gzip
import json
import os

# Support for both Python2.X and 3.X.
# -----------------------------------------------------------------------------
try:
    from io import StringIO
    from urllib.request import urlopen
    from urllib.error import HTTPError
    from urllib.parse import quote_plus
except ImportError:
    from StringIO import StringIO
    from urllib2 import urlopen, HTTPError
    from urllib import quote_plus

try:
    input_shim = raw_input
except NameError:
    # If `raw_input` throws a `NameError`, the user is using Python 2.X.
    input_shim = input


# Enumerables and constants
# -----------------------------------------------------------------------------

class Enum(set):
    """Simple Enum shim since Python 2.X does not have them.
    """

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError


# The entity types supported by the Harmonizome API.
class Entity(Enum):

    DATASET = 'dataset'
    GENE = 'gene'
    GENE_SET = 'gene_set'
    ATTRIBUTE = 'attribute'
    GENE_FAMILY = 'gene_family'
    NAMING_AUTHORITY = 'naming_authority'
    PROTEIN = 'protein'
    RESOURCE = 'resource'


def json_from_url(url):
    """Returns API response after decoding and loading JSON.
    """
    response = urlopen(url)
    data = response.read().decode('utf-8')
    return json.loads(data)


VERSION = '1.0'
API_URL = 'http://amp.pharm.mssm.edu/Harmonizome/api'
DOWNLOAD_URL = 'http://amp.pharm.mssm.edu/static/hdfs/harmonizome/data'

# This config objects pulls the names of the datasets, their directories, and
# the possible downloads from the API. This allows us to add new datasets and
# downloads without breaking this file.
try:
    config = json_from_url(API_URL + '/dark/script_config')
    DOWNLOADS = [x for x in config.get('downloads')]
    DATASET_TO_PATH = config.get('datasets')
except:
    DOWNLOADS = []
    DATASET_TO_PATH = {}

# Harmonizome class
# -----------------------------------------------------------------------------

class Harmonizome(object):

    __version__ = VERSION
    DATASETS = DATASET_TO_PATH.keys()

    @classmethod
    def get(cls, entity, name=None, start_at=None):
        """Returns a single entity or a list, depending on if a name is
        provided. If no name is provided and start_at is specified, returns a
        list starting at that cursor position.
        """
        if name:
            name = quote_plus(name)
            return _get_by_name(entity, name)
        if start_at is not None and type(start_at) is int:
            return _get_with_cursor(entity, start_at)
        url = '%s/%s/%s' % (API_URL, VERSION, entity)
        result = json_from_url(url)
        return result

    @classmethod
    def next(cls, response):
        """Returns the next set of entities based on a previous API response.
        """
        start_at = _get_next(response)
        entity = _get_entity(response)
        return cls.get(entity=entity, start_at=start_at)

    @classmethod
    def download(cls, datasets=None):
        """For each dataset, creates a directory and downloads files into it.
        """
        # Why not check `if not datasets`? Because in principle, a user could 
        # call `download([])`, which should download nothing, not everything.
        # Why might they do this? Imagine that the list of datasets is
        # dynamically generated in another user script.
        if datasets is None:
            datasets = cls.DATASETS
            warning = 'Warning: You are going to download all Harmonizome '\
                      'data. This is roughly 30GB. Do you accept?\n(Y/N) '
            resp = input_shim(warning)
            if resp.lower() != 'y':
                return

        for dataset in datasets:
            if dataset not in cls.DATASETS:
                msg = '"%s" is not a valid dataset name. Check the `DATASETS`'\
                      ' property for a complete list of names.' % dataset
                raise AttributeError(msg)
            if not os.path.exists(dataset):
                os.mkdir(dataset)

            for dl in DOWNLOADS:
                path = DATASET_TO_PATH[dataset]
                url = '%s/%s/%s' % (DOWNLOAD_URL, path, dl)

                try:
                    response = urlopen(url)
                except HTTPError:
                    # Not every dataset has all downloads.
                    pass

                filename = '%s/%s' % (dataset, dl)
                filename = filename.replace('.gz', '')

                if response.code != 200:
                    raise Exception('This should not happen')

                _download_and_decompress_file(response, filename)


# Utility functions
# -------------------------------------------------------------------------

def _get_with_cursor(entity, start_at):
    """Returns a list of entities based on cursor position.
    """
    url = '%s/%s/%s?cursor=%s' % (API_URL, VERSION, entity,str(start_at))
    result = json_from_url(url)
    return result


def _get_by_name(entity, name):
    """Returns a single entity based on name.
    """
    url = '%s/%s/%s/%s' % (API_URL, VERSION, entity, name)
    return json_from_url(url)


def _get_entity(response):
    """Returns the entity from an API response.
    """
    path = response['next'].split('?')[0]
    return path.split('/')[3]


def _get_next(response):
    """Returns the next property from an API response.
    """
    if response['next']:
        return int(response['next'].split('=')[1])
    return None


# This function was adopted from here: http://stackoverflow.com/a/15353312.
def _download_and_decompress_file(response, filename):
    """Downloads and decompresses a single file from a response object.
    """
    compressed_file = StringIO()
    compressed_file.write(response.read())
    compressed_file.seek(0)
    decompressed_file = gzip.GzipFile(fileobj=compressed_file, mode='rb')
    with open(filename, 'w+') as outfile:
        outfile.write(decompressed_file.read())


def _download_and_decompress_file(response, filename):
    """
    """
    compressed_file = io.BytesIO(response.read())
    decompressed_file = gzip.GzipFile(fileobj=compressed_file)

    with open(filename, 'wb+') as outfile:
        outfile.write(decompressed_file.read())




import json
import urllib.request

from bioservices import ChEBI
from bioservices import EUtils
from bioservices import Ensembl
from bioservices import Reactome
from bioservices import UniProt
from bioservices.kegg import KEGG
from loguru import logger
from tqdm import tqdm


################################################################################
### Gene-related functions                                                   ###
################################################################################


def batch(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def get_ensembl_metadata_online(ensembl_ids):

    ensembl_ids = list(set(ensembl_ids))
    logger.info('get_ensembl_metadata %d' % len(ensembl_ids))

    BATCH_SIZE = 1000
    ens = Ensembl()
    ensembl_lookup = {}
    cumulative_total = 0
    for x in batch(ensembl_ids, BATCH_SIZE):
        batch_ids = [i for i in x]
        cumulative_total += len(batch_ids)
        logger.debug('%d/%d' % (cumulative_total, len(ensembl_ids)))
        lookup = ens.post_lookup_by_id(identifiers=batch_ids)
        ensembl_lookup.update(lookup)

    return ensembl_lookup


def get_single_ensembl_metadata_online(ensembl_id):
    ens = Ensembl()
    res = ens.get_lookup_by_id(ensembl_id, expand=True)
    return res


def get_entrez_summary(gene_id):
    e = EUtils()
    ret = e.ESearch('gene', gene_id)
    ret = e.EFetch('gene', ret['idlist'])
    return ret.decode("utf-8")


def get_gene_names(ensembl_ids, gtf_dict):
    metadata_map = {}
    for ensembl_id in ensembl_ids:
        try:
            display_name = gtf_dict[ensembl_id]
            display_name = clean_label(display_name)
            metadata_map[ensembl_id] = {'display_name': display_name}
        except KeyError:
            metadata_map[ensembl_id] = {'display_name': ensembl_id}
    return metadata_map


################################################################################
### Protein-related functions                                                ###
################################################################################


def get_uniprot_metadata_online(uniprot_ids):

    uniprot_ids = list(set(uniprot_ids))
    logger.info('get_uniprot_metadata %d' % len(uniprot_ids))

    BATCH_SIZE = 200
    uniprot = UniProt()
    uniprot_lookup = {}

    cumulative_total = 0
    for x in batch(uniprot_ids, BATCH_SIZE):
        batch_ids = [i for i in x]
        cumulative_total += len(batch_ids)
        logger.debug('%d/%d' % (cumulative_total, len(uniprot_ids)))

        res = uniprot.retrieve(batch_ids)
        for r in res:
            for key in r['accession']:
                protein_id = key.contents[0]
                for x in r['recommendedname']:
                    tag = x.find('shortname')
                    if tag is None:
                        tag = x.find('fullname')
                    label = tag.contents[0]
                    uniprot_lookup[protein_id] = {'display_name': label}

    return uniprot_lookup


def get_single_uniprot_metadata_online(uniprot_id):
    uniprot = UniProt()
    res = uniprot.retrieve(uniprot_id)
    return res


def clean_label(w):
    results = []
    for tok in w.split(' '):
        # filtered = re.sub(r'[^\w\s]', '', tok)
        filtered = tok.replace("'", "")
        filtered = filtered.replace('"', "")
        results.append(filtered.strip())
    return ' '.join(results)


def get_uniprot_metadata_reactome(uniprot_ids):
    protein_descriptions = {}  # TODO: get the description of each protein from reactome
    metadata_map = {}
    for protein_id in protein_descriptions:
        metadata_map[protein_id] = {'display_name': protein_id}
        if protein_id in protein_descriptions and protein_descriptions[protein_id] is not None:
            desc = protein_descriptions[protein_id]
            tokens = desc.split(':')
            for i in range(len(tokens)):
                w = tokens[i]
                if w.startswith('recommendedName'):
                    next_w = clean_label(tokens[i + 1])
                    metadata_map[protein_id] = {'display_name': next_w}
                    logger.debug('%d -- %s' % (protein_id, next_w))
    return metadata_map


################################################################################
### Compound-related functions                                               ###
################################################################################

def kegg_to_chebi(compound_ids):
    results = {}
    ch = ChEBI()
    for compound_id in compound_ids:
        try:
            if compound_id.startswith('C'):
                res = ch.getLiteEntity(compound_id)
                assert len(res) > 0 # we should always be able to convert from KEGG -> ChEBI ids
                le = res[0]
                chebi_number = le.chebiId.split(':')[1]
                logger.debug('KEGG %s --> ChEBI %s' % (compound_id, chebi_number))
                results[compound_id] = chebi_number
            else:
                logger.debug('ChEBI %s --> ChEBI %s' % (compound_id, compound_id))
                results[compound_id] = compound_id
        except Exception:
            logger.debug('KEGG %s --> KEGG %s' % (compound_id, compound_id))
            results[compound_id] = compound_id
    return results


def get_compound_metadata_online(kegg_ids):
    s = KEGG()
    metadata_map = {}
    my_list = list(range(len(kegg_ids)))
    failed = 0
    for i in tqdm(my_list):
        try:
            kegg_id = kegg_ids[i]
            res = s.get(kegg_id)
            d = s.parse(res)
            first_name = d['NAME'][0]
            first_name = first_name.replace(';', '') # strip last ';' character
            metadata_map[kegg_id] = {'display_name': first_name}
        except TypeError:
            failed += 1
    logger.info('Failed = %d' % failed)

    return metadata_map


def get_compound_metadata(compound_ids, kegg_id_to_display_names, id_to_names):
    metadata_map = {}
    for compound_id in compound_ids:
        try:
            display_name = clean_label(kegg_id_to_display_names[compound_id]['display_name'])
        except KeyError:
            try:
                display_name = clean_label(id_to_names[compound_id])
                idx = display_name.find('[ChEBI') # remove [ChEBI ...]
                display_name = display_name[0:idx].strip()
            except KeyError:
                display_name = compound_id
        metadata_map[compound_id] = {'display_name': display_name}
    return metadata_map


def get_single_compound_metadata_online(compound_id):

    if compound_id.upper().startswith('C'):
        s = KEGG()
        res = s.get(compound_id)
        return s.parse(res)
    else:
        ch = ChEBI()
        res = ch.getCompleteEntity('CHEBI:'+compound_id)
        return res


################################################################################
### Reaction-related functions                                               ###
################################################################################


def get_reactome_content_service(reactome_id):
    json_url = 'https://reactome.org/ContentService/data/query/' + reactome_id
    with urllib.request.urlopen(json_url) as url:
        lookup = json.loads(url.read().decode())
        return lookup


################################################################################
### Pathway-related functions                                                ###
################################################################################


################################################################################
### Species-related functions                                                ###
################################################################################


def get_species_name_to_id():
    r = Reactome()
    species_name_to_id = {x['displayName']: x['dbId'] for x in r.data_species_all()}
    return species_name_to_id


from collections import defaultdict
import os

import pandas as pd
import xmltodict
from bioservices.kegg import KEGG
from loguru import logger
from neo4j import GraphDatabase, basic_auth


def get_neo4j_driver():
    NEO4J_SERVER = os.getenv('NEO4J_SERVER', 'bolt://localhost:7687')
    if 'NEO4J_SERVER' not in os.environ:
        logger.warning('Using a default neo4j server: %s' % NEO4J_SERVER)

    NEO4J_USER = os.getenv('NEO4J_USER', 'neo4j')
    NEO4J_PASSWORD = os.getenv('NEO4J_PASSWORD', 'neo4j')
    if 'NEO4J_USER' not in os.environ or 'NEO4J_PASSWORD' not in os.environ:
        logger.warning('Using a default neo4j username or password: %s' % NEO4J_USER)

    try:
        neo4j_driver = GraphDatabase.driver(NEO4J_SERVER,
                                            auth=basic_auth(NEO4J_USER, NEO4J_PASSWORD))
        logger.info('Created graph database driver for %s (%s)' % (NEO4J_SERVER, NEO4J_USER))
        return neo4j_driver
    except Exception as e:
        logger.warning('Failed to connect to graph database: %s' % str(e))
        raise e


driver = get_neo4j_driver()


def get_neo4j_session():
    session = None
    try:
        session = driver.session()
    except Exception:
        raise
    return session


def get_species_list():
    results = []
    try:
        session = get_neo4j_session()
        query = """
        MATCH (n:Species) RETURN n.displayName AS name order by name        
        """
        query_res = session.run(query)
        logger.debug(query)
        for record in query_res:
            results.append(record['name'])
    finally:
        if session is not None: session.close()
    return results


def get_species_dict():
    species_list = get_species_list()
    species_dict = {}
    for idx, s in enumerate(species_list):
        species_dict[str(idx)] = s
    return species_dict


################################################################################
### Gene-related functions                                                   ###
################################################################################


def ensembl_to_uniprot(ensembl_ids, species_list):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()
        query = """
        MATCH
            (rg:ReferenceGeneProduct)-[:referenceGene]->
            (rs:ReferenceSequence)-[:species]->(s:Species)
        WHERE
            rs.identifier IN {ensembl_ids} AND
            rs.databaseName = 'ENSEMBL' AND            
            s.displayName IN {species}
        RETURN DISTINCT
            rs.identifier AS gene_id,
            rs.databaseName AS gene_db,
            rg.identifier AS protein_id,
            rg.databaseName AS protein_db,
            rg.url as URL
        """
        params = {
            'ensembl_ids': ensembl_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            gene_id = record['gene_id']
            protein_id = record['protein_id']
            results[gene_id].append(protein_id)

    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


################################################################################
### Protein-related functions                                                ###
################################################################################


def uniprot_to_ensembl(uniprot_ids, species_list):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()
        query = """
        MATCH
            (rg:ReferenceGeneProduct)-[:referenceGene]->
            (rs:ReferenceSequence)-[:species]->(s:Species)
        WHERE
            rg.identifier IN {uniprot_ids} AND
            rs.databaseName = 'ENSEMBL' AND
            s.displayName IN {species}
        RETURN DISTINCT
            rs.identifier AS gene_id,
            rs.databaseName AS gene_db,
            rg.identifier AS protein_id,
            rg.databaseName AS protein_db,
            rg.url as URL
        """
        params = {
            'uniprot_ids': uniprot_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            gene_id = record['gene_id']
            protein_id = record['protein_id']
            results[protein_id].append(gene_id)

    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


def uniprot_to_reaction(uniprot_ids, species_list):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()

        # note that using hasComponent|hasMember|hasCandidate below will
        # retrieve all the sub-complexes too
        query = """
        MATCH (rle:ReactionLikeEvent)-[:input|output|catalystActivity
              |physicalEntity|regulatedBy|regulator|hasComponent|hasMember
              |hasCandidate*]->
              (pe:PhysicalEntity)-[:referenceEntity]->
              (re:ReferenceEntity)-[:referenceDatabase]->
              (rd:ReferenceDatabase)
        WHERE
            re.identifier IN {uniprot_ids} AND
            rd.displayName = 'UniProt' AND
            rle.speciesName IN {species}
        RETURN DISTINCT
            re.identifier AS protein_id,
            re.description AS description,
            rd.displayName AS protein_db,
            rle.stId AS reaction_id,
            rle.displayName AS reaction_name
        """
        params = {
            'uniprot_ids': uniprot_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            protein_id = record['protein_id']
            item = {
                'reaction_id': record['reaction_id'],
                'reaction_name': record['reaction_name']
            }
            results[protein_id].append(item)
    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


################################################################################
### Compound-related functions                                               ###
################################################################################


def get_all_compound_ids():
    results = []
    try:
        session = get_neo4j_session()
        query = """
        MATCH (di:DatabaseIdentifier)
        WHERE
            di.databaseName = 'COMPOUND'
        RETURN DISTINCT
            di.displayName AS compound_id
        """
        query_res = session.run(query)
        logger.debug(query)

        for record in query_res:
            key = record['compound_id'].split(':')  # e.g. 'COMPOUND:C00025'
            compound_id = key[1]
            results.append(compound_id)
    finally:
        if session is not None: session.close()
    return results


def compound_to_reaction(compound_ids, species_list):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()
        query = """
        MATCH (rle:ReactionLikeEvent)-[:input|output|catalystActivity
              |physicalEntity|regulatedBy|regulator|hasComponent|hasMember
              |hasCandidate*]->
              (pe:PhysicalEntity)-[:crossReference|:referenceEntity]->
              (do:DatabaseObject)
        WHERE
            do.identifier IN {compound_ids} AND
            rle.speciesName IN {species}
        RETURN DISTINCT
            do.identifier AS compound_id,
            do.displayName as display_name,
            do.databaseName AS compound_db,
            rle.stId AS reaction_id,
        	rle.displayName AS reaction_name        
        """
        params = {
            'compound_ids': compound_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            compound_id = record['compound_id']
            item = {
                'reaction_id': record['reaction_id'],
                'reaction_name': record['reaction_name']
            }
            results[compound_id].append(item)
            compound_name = record['display_name']
            id_to_names[compound_id] = compound_name
    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


def produce_kegg_dict(kegg_location, param):
    with open(kegg_location) as kegg_cmpd_file:
        cmpd_dict = xmltodict.parse(kegg_cmpd_file.read())

    kegg_dict = {}
    for compound in cmpd_dict['compounds']['compound']:
        kegg_dict[compound[param]] = compound['formula']

    return kegg_dict


################################################################################
### Reaction-related functions                                               ###
################################################################################


# get all the entities involved in a reaction
def get_reaction_entities(reaction_ids):
    results = defaultdict(list)
    try:
        session = get_neo4j_session()
        query = """
        MATCH (rle:ReactionLikeEvent)-[rr:input|output|catalystActivity
              |physicalEntity|regulatedBy|regulator|hasComponent|hasMember
              |hasCandidate*]->(dbo:DatabaseObject)
        WHERE
            rle.stId IN {reaction_ids}
        RETURN
            rle.stId AS reaction_id,
            dbo.stId AS entity_id,
            dbo.schemaClass AS schema_class,
            dbo.displayName as display_name,
            extract(rel IN rr | type(rel)) AS types
        """
        params = {
            'reaction_ids': reaction_ids
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            reaction_id = record['reaction_id']
            entity_id = record['entity_id']
            schema_class = record['schema_class']
            display_name = record['display_name']
            relationship_types = record['types']
            item = (schema_class, entity_id, display_name, relationship_types)
            results[reaction_id].append(item)
    finally:
        if session is not None: session.close()
    return results


def reaction_to_uniprot(reaction_ids, species_list):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()

        # note that using hasComponent|hasMember|hasCandidate below will
        # retrieve all the sub-complexes too
        query = """
        MATCH (rle:ReactionLikeEvent)-[:input|output|catalystActivity
              |physicalEntity|regulatedBy|regulator|hasComponent|hasMember
              |hasCandidate*]->
              (pe:PhysicalEntity)-[:referenceEntity]->
              (re:ReferenceEntity)-[:referenceDatabase]->
              (rd:ReferenceDatabase)
        WHERE
            rle.stId IN {reaction_ids} AND
            rd.displayName = 'UniProt' AND
            rle.speciesName IN {species}
        RETURN DISTINCT
            re.identifier AS protein_id,
            re.description AS description,
            rd.displayName AS protein_db,
            rle.stId AS reaction_id,
            rle.displayName AS reaction_name
        """
        params = {
            'reaction_ids': reaction_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            protein_id = record['protein_id']
            reaction_id = record['reaction_id']
            results[reaction_id].append(protein_id)
    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


def reaction_to_compound(reaction_ids, species_list, use_kegg=False):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()
        query = """
        MATCH (rle:ReactionLikeEvent)-[:input|output|catalystActivity
              |physicalEntity|regulatedBy|regulator|hasComponent|hasMember
              |hasCandidate*]->
              (pe:PhysicalEntity)-[:crossReference|:referenceEntity]->
              (do:DatabaseObject)
        WHERE
            rle.stId IN {reaction_ids} AND        
            (do.databaseName = 'COMPOUND' OR do.databaseName = 'ChEBI') AND
            rle.speciesName IN {species}
        RETURN DISTINCT
            do.identifier as compound_id,
            do.displayName as display_name,            
            do.databaseName AS compound_db,
            rle.stId AS reaction_id,
        	rle.displayName AS reaction_name
        """
        params = {
            'reaction_ids': reaction_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            reaction_id = record['reaction_id']
            compound_id = record['compound_id']
            display_name = record['display_name']
            database_name = record['compound_db']
            # TODO: find better ways to remove duplicates between KEGG and ChEBI?
            valid = False
            if use_kegg:
                if database_name == 'COMPOUND':
                    valid = True
            else:
                if database_name == 'ChEBI':
                    valid = True
            if valid:
                results[reaction_id].append(compound_id)
                id_to_names[compound_id] = display_name
    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


def rchop(thestring, ending):
    if thestring.endswith(ending):
        return thestring[:-len(ending)]
    return thestring


def reaction_to_pathway(reaction_ids, species_list, metabolic_pathway_only, leaf=True):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()

        # initial match clause in the query
        query = """
        MATCH (tp:TopLevelPathway)-[:hasEvent*]->
              (p:Pathway)-[:hasEvent*]->(rle:ReactionLikeEvent)
        WHERE
            tp.speciesName IN {species} AND        
            rle.stId IN {reaction_ids} AND            
        """

        if leaf:  # retrieve only the leaf nodes in the pathway hierarchy
            query += " (p)-[:hasEvent]->(rle) AND "

        if metabolic_pathway_only:  # only retrieves metabolic pathways
            query += " tp.displayName = 'Metabolism' AND "

        # remove last AND
        query = rchop(query.strip(), 'AND')

        # add return clause
        query += """
        RETURN
            rle.stId AS reaction_id,
            rle.displayName AS reaction_name,
            rle.speciesName AS reaction_species,
            p.stId AS pathway_id,
            p.displayName AS pathway_name,
            tp.speciesName AS pathway_species
        """

        params = {
            'reaction_ids': reaction_ids,
            'species': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            reaction_id = record['reaction_id']
            reaction_name = record['reaction_name']
            reaction_species = record['reaction_species']
            pathway_id = record['pathway_id']
            pathway_name = record['pathway_name']
            pathway_species = record['pathway_species']
            item = {
                'pathway_id': pathway_id,
                'pathway_name': pathway_name
            }
            results[reaction_id].append(item)
            id_to_names[reaction_id] = {'name': reaction_name, 'species': reaction_species}
            id_to_names[pathway_id] = {'name': pathway_name, 'species': pathway_species}
    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


################################################################################
### Pathway-related functions                                                ###
################################################################################


def pathway_to_reactions(pathway_ids):
    id_to_names = {}
    results = defaultdict(list)
    try:
        session = get_neo4j_session()
        # retrieve only the leaf nodes in the pathway hierarchy
        query = """
        MATCH (p:Pathway)-[:hasEvent*]->(rle:ReactionLikeEvent)
        WHERE
            p.stId IN {pathway_ids} AND
            (p)-[:hasEvent]->(rle)
        RETURN
            rle.stId AS reaction_id,
            rle.displayName AS reaction_name,
            p.stId AS pathway_id,
            p.displayName AS pathway_name
        """
        params = {
            'pathway_ids': pathway_ids
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            reaction_id = record['reaction_id']
            reaction_name = record['reaction_name']
            pathway_id = record['pathway_id']
            pathway_name = record['pathway_name']
            results[pathway_id].append(reaction_id)
            id_to_names[reaction_id] = reaction_name
            id_to_names[pathway_id] = pathway_name
    finally:
        if session is not None: session.close()
    return dict(results), id_to_names


def get_reactome_description(reactome_id, from_parent=False):
    results = []
    try:
        session = get_neo4j_session()
        if from_parent:
            query = """
            MATCH (dbo1:DatabaseObject)<-[:inferredTo*]-(dbo2:DatabaseObject)-[:summation|:literatureReference]-(ss)
                    WHERE
                        dbo1.stId = {reactome_id} AND
                        dbo2.isInferred = False
                    RETURN
                        dbo2.stId as reactome_id,
                        dbo2.speciesName as species,
                        dbo2.isInferred as inferred,
                        ss.displayName as display_name,
                        ss.text as summary_text,
                        ss.schemaClass as summary_type,
                        properties(ss) as summary
            """
        else:
            query = """
            MATCH (dbo:DatabaseObject)-[:summation|:literatureReference]-(ss)
                    WHERE
                        dbo.stId = {reactome_id}
                    RETURN
                        dbo.stId as reactome_id,
                        dbo.speciesName as species,
                        dbo.isInferred as inferred,
                        ss.displayName as display_name,
                        ss.text as summary_text,
                        ss.schemaClass as summary_type,
                        properties(ss) as summary_props
            """
        params = {
            'reactome_id': reactome_id,
        }
        query_res = session.run(query, params)
        logger.debug(query)
        record_list = list(query_res.records())
        results = list(map(lambda x: x.data(), record_list))
        first_data = results[0]
        is_inferred = first_data['inferred']
    finally:
        if session is not None: session.close()
    return results, is_inferred


def retrieve_kegg_formula(reactome_compound_name):
    k = KEGG()
    compound_name = reactome_compound_name.replace('COMPOUND', 'cpd')
    res = k.get(compound_name).split('\n')
    for line in res:
        if line.startswith('FORMULA'):
            formula = line.split()[1]  # get the second token
            return formula
    return None


def get_all_pathways(species_list):
    results = []
    try:
        session = get_neo4j_session()

        # retrieve only the leaf nodes in the pathway hierarchy
        query = """
            MATCH (tp:TopLevelPathway)-[:hasEvent*]->(p:Pathway)-[:hasEvent*]->(rle:ReactionLikeEvent)
            WHERE
                tp.displayName = 'Metabolism' AND
                tp.speciesName IN {species_list} AND
                (p)-[:hasEvent]->(rle)
            RETURN DISTINCT
                p.speciesName AS species_name,            
                p.displayName AS pathway_name,
                p.stId AS pathway_id                       
            ORDER BY species_name, pathway_name
        """
        params = {
            'species_list': species_list
        }
        query_res = session.run(query, params)
        logger.debug(query)

        for record in query_res:
            pathway_species = record['species_name']
            pathway_name = record['pathway_name']
            pathway_id = record['pathway_id']
            results.append((pathway_species, pathway_name, pathway_id))
    finally:
        if session is not None: session.close()
    return results


def get_all_pathways_formulae(species):
    results = defaultdict(set)
    pathway_id_to_name = {}
    try:
        session = get_neo4j_session()

        # retrieve only the leaf nodes in the pathway hierarchy
        query = """
        MATCH (tp:TopLevelPathway)-[:hasEvent*]->
              (p:Pathway)-[:hasEvent*]->(rle:ReactionLikeEvent),
              (rle)-[:input|output|catalystActivity|physicalEntity|regulatedBy|regulator|hasComponent
              |hasMember|hasCandidate*]->(pe:PhysicalEntity),
              (pe:PhysicalEntity)-[:crossReference]->(di:DatabaseIdentifier)<-[:crossReference]-(rm:ReferenceMolecule)
        WHERE
              tp.displayName = 'Metabolism' AND
              tp.speciesName = {species} AND
              di.databaseName = 'COMPOUND' AND
              (p)-[:hasEvent]->(rle)
        RETURN DISTINCT
            p.schemaClass,
            p.displayName AS pathway_name,
            p.stId AS pathway_id,
            di.displayName as compound_name,
            rm.formula AS formula,
            di.url
        """
        params = {
            'species': species
        }
        query_res = session.run(query, params)
        logger.debug(query)

        i = 0
        retrieved = {}
        for record in query_res:
            pathway_id = record['pathway_id']
            pathway_name = record['pathway_name']
            pathway_id_to_name[pathway_id] = pathway_name
            compound_name = record['compound_name']
            formula = record['formula']
            if formula is None:
                if compound_name not in retrieved:
                    formula = retrieve_kegg_formula(compound_name)
                    logger.debug('Missing formula for %s, retrieved %s from kegg' %
                                 (compound_name, formula))
                    retrieved[compound_name] = formula
                else:
                    formula = retrieved[compound_name]
            assert formula is not None, 'Formula is missing for %s' % compound_name
            results[pathway_id].add(formula)
    finally:
        if session is not None: session.close()
    return dict(results), pathway_id_to_name


################################################################################
### Analysis functions                                                       ###
################################################################################


def get_reaction_ids(mapping):
    all_reactions = []
    for key in mapping:
        rs = mapping[key]
        rids = [r['reaction_id'] for r in rs]
        all_reactions.extend(rids)
    return all_reactions


def get_reactions_from_mapping(mapping):
    reaction_names = {}
    reaction_members = defaultdict(list)
    for key in mapping:
        for reaction in mapping[key]:
            r_id = reaction['reaction_id']
            r_name = reaction['reaction_name']
            reaction_names[r_id] = r_name
            reaction_members[r_id].append(key)
    assert reaction_names.keys() == reaction_members.keys()
    return reaction_names, dict(reaction_members)


def get_protein_to_gene(mapping):
    protein_to_gene = defaultdict(list)
    for gene_id in mapping:
        for protein_id in mapping[gene_id]:
            protein_to_gene[protein_id].append(gene_id)
    return dict(protein_to_gene)


def merge_two_dicts(x, y):
    # https://stackoverflow.com/questions/38987/how-to-merge-two-dictionaries-in-a-single-expression
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z


def get_coverage(observed_count, total_count):
    try:
        return observed_count / float(total_count)
    except ZeroDivisionError:
        return 0


def get_reaction_df(transcript_mapping, protein_mapping, compound_mapping,
                    pathway_mapping, species_list):
    r_name_1, r_members_1 = get_reactions_from_mapping(protein_mapping)
    r_name_2, r_members_2 = get_reactions_from_mapping(compound_mapping)
    reaction_names = merge_two_dicts(r_name_1, r_name_2)
    protein_to_gene = get_protein_to_gene(transcript_mapping)

    pathway_compound_counts = defaultdict(int)
    pathway_protein_counts = defaultdict(int)

    reaction_ids = set(list(r_members_1.keys()) + list(r_members_2.keys()))
    reaction_entities = get_reaction_entities(list(reaction_ids))
    rows = []
    for reaction_id in reaction_ids:

        observed_protein_count = 0
        observed_compound_count = 0
        protein_str = ''
        compound_str = ''

        if reaction_id in r_members_1:
            proteins = r_members_1[reaction_id]
            observed_protein_count = len(proteins)
            for prot in proteins:
                if prot in protein_to_gene:
                    protein_str += '%s (%s):' % (prot, ':'.join(
                        protein_to_gene[prot]))
                else:
                    protein_str += '%s:' % prot
            protein_str = protein_str.rstrip(':')  # remove last :

        if reaction_id in r_members_2:
            compounds = r_members_2[reaction_id]
            observed_compound_count = len(compounds)
            compound_str = ':'.join(compounds)

        entities = reaction_entities[reaction_id]
        all_compound_count = len([x for x in entities
                                  if x[0] == 'SimpleEntity'])
        all_protein_count = len([x for x in entities
                                 if x[0] == 'EntityWithAccessionedSequence'])

        reaction_name = reaction_names[reaction_id]
        protein_coverage = get_coverage(observed_protein_count,
                                        all_protein_count)
        compound_coverage = get_coverage(observed_compound_count,
                                         all_compound_count)
        s1 = observed_protein_count + observed_compound_count
        s2 = all_protein_count + all_compound_count
        all_coverage = get_coverage(s1, s2)

        if reaction_id in pathway_mapping:
            pathway_id_str = ':'.join([x['pathway_id']
                                       for x in pathway_mapping[reaction_id]])
            pathway_name_str = ':'.join([x['pathway_name']
                                         for x in pathway_mapping[reaction_id]])
            protein_coverage_str = '%.2f' % protein_coverage
            compound_coverage_str = '%.2f' % compound_coverage
            all_coverage_str = '%.2f' % all_coverage
            row = (reaction_id,
                   reaction_name,
                   protein_coverage_str,
                   compound_coverage_str,
                   all_coverage_str,
                   protein_str,
                   observed_protein_count,
                   all_protein_count,
                   compound_str,
                   observed_compound_count,
                   all_compound_count,
                   pathway_id_str,
                   pathway_name_str)
            rows.append(row)

            # accumulate the count for pathways too
            for x in pathway_mapping[reaction_id]:
                pathway_compound_counts[x['pathway_id']] += observed_compound_count
                pathway_protein_counts[x['pathway_id']] += observed_protein_count

    df = pd.DataFrame(rows, columns=['reaction_id',
                                     'reaction_name',
                                     'protein_coverage',
                                     'compound_coverage',
                                     'all_coverage',
                                     'protein',
                                     'observed_protein_count',
                                     'all_protein_count',
                                     'compound',
                                     'observed_compound_count',
                                     'all_compound_count',
                                     'pathway_ids',
                                     'pathway_names'])
    return df, pathway_compound_counts, pathway_protein_counts



import numpy as np
import pandas as pd
from loguru import logger

from .common import as_list
from .constants import REACTIONS, PATHWAYS, MAPPING, GENES, PROTEINS, COMPOUNDS, TRANSCRIPTS, QUERY_DISPLAY_NAME, \
    QUERY_NODE_ID, QUERY_DATA_TYPE, QUERY_OBSERVED, QUERY_ENTITY_ID, QUERY_SOURCE_ID
from .info import get_info


class QueryBuilder():
    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.tasks = []
        self.results = []

    def add(self, task):
        self.tasks.append(task)
        return self

    def run(self):
        self.results = []
        prev_task = None
        for task in self.tasks:
            # run task and get its result
            task.run(self.pipeline, prev_task)
            res = task.get_result()

            # track all results generated so far
            self.results.append(res)
            prev_task = task

        # return the last result, if any
        if len(self.results) > 0:
            return self.results[-1]
        else:
            return None


class Query():
    def __init__(self):
        self.result = None

    def get_result(self):
        return self.result

    def run(self, pipeline, previous_query):
        raise NotImplementedError

    def _get_node_data(self, mapping, node_id):
        node_info = mapping.get_node(node_id)
        display_name = node_info[QUERY_DISPLAY_NAME]
        obs = mapping.is_observed(node_id)
        data_type = mapping.get_node_type(node_id)
        return {
            QUERY_NODE_ID: node_id,
            QUERY_DISPLAY_NAME: display_name,
            QUERY_DATA_TYPE: data_type,
            QUERY_OBSERVED: obs
        }

    def _to_row(self, node_data):
        node_id = node_data[QUERY_NODE_ID]
        display_name = node_data[QUERY_DISPLAY_NAME]
        data_type = node_data[QUERY_DATA_TYPE]
        obs = node_data[QUERY_OBSERVED]
        row = [node_id, display_name, MAPPING[data_type], obs]
        return row

    def _to_df(self, data):
        columns = [QUERY_ENTITY_ID, QUERY_DISPLAY_NAME, QUERY_DATA_TYPE, QUERY_OBSERVED]
        df = pd.DataFrame(data, columns=columns)
        df = df.set_index(QUERY_ENTITY_ID)
        return df


class Entity(Query):
    def __init__(self, node_ids):
        super().__init__()
        self.node_ids = as_list(node_ids)

    def run(self, pipeline, _):
        mapping = pipeline.mapping
        data = []
        for node_id in self.node_ids:
            try:
                node_data = self._get_node_data(mapping, node_id)
                row = self._to_row(node_data)
                data.append(row)
            except KeyError:
                logger.warning(" %s wasn't mapped, it may not be present in Reactome" % node_id)
        df = self._to_df(data)
        self.result = df


class Select(Query):
    def __init__(self, data_types):
        super().__init__()
        self.data_types = as_list(data_types)

    def run(self, pipeline, previous_query):
        mapping = pipeline.mapping
        # get all nodes of data_type from the mapping object
        res = mapping.get_nodes(types=self.data_types)
        node_ids = [node_id for node_id, node_data in res]

        # select entities having node_ids, and save the result
        e = Entity(node_ids)
        e.run(pipeline, previous_query)
        self.result = e.get_result()


class Connected(Query):
    def __init__(self, data_type=None, observed=None):
        super().__init__()
        self.data_types = as_list(data_type) if data_type is not None else [GENES, TRANSCRIPTS,
                                                                            PROTEINS, COMPOUNDS,
                                                                            REACTIONS, PATHWAYS]
        self.observed = observed

    def run(self, pipeline, previous_query):
        mapping = pipeline.mapping
        previous_res = previous_query.get_result()
        node_ids = previous_res.index.values
        possible_obs = as_list(self.observed) if self.observed is not None else [True, False]

        data = []
        for node_id in node_ids:
            for data_type in self.data_types:
                connected = self._search_graph(mapping, node_id, data_type)
                for connected_id in connected:
                    if connected_id == node_id:
                        continue

                    node_data = self._get_node_data(mapping, connected_id)
                    row = self._to_row(node_data) + [node_id]

                    obs = node_data['observed']
                    if obs is None:  # reactions, pathways don't have any observed value
                        data.append(row)
                    elif obs in possible_obs:  # everything else
                        data.append(row)

        df = self._to_df(data).drop_duplicates()
        self.result = df

    def _to_df(self, data):
        columns = [QUERY_ENTITY_ID, QUERY_DISPLAY_NAME, QUERY_DATA_TYPE, QUERY_OBSERVED, QUERY_SOURCE_ID]
        df = pd.DataFrame(data, columns=columns)
        df = df.set_index(QUERY_ENTITY_ID)
        return df

    def _search_graph(self, mapping, node_id, dest_type):
        visited = []  # list of visited nodes
        queue = []  # nodes to process next
        seen_types = set()  # omics type seen so far

        visited.append(node_id)
        queue.append(node_id)

        results = []
        while queue:
            s = queue.pop(0)
            node_type = mapping.get_node_type(s)
            seen_types.add(node_type)

            if node_type == dest_type:
                # terminate search for s
                results.append(s)
                continue
            else:
                # keep searching in the neighbours of s
                # but exclude nodes of type that we've seen before
                neighbours = mapping.get_neighbours(s, exclude_types=seen_types)
                for node in neighbours:
                    if node not in visited:
                        visited.append(node)
                        queue.append(node)
        return results


class SignificantDE(Query):
    def __init__(self, case, control, pval, fc_lte=None, fc_gte=None, N=None, ascending=None):
        self.case = case
        self.control = control
        self.pval = pval
        self.fc_lte = fc_lte
        self.fc_gte = fc_gte
        self.N = N
        self.ascending = ascending

    def run(self, pipeline, previous_query):
        # get data type from the previous query, ensure it's only for one type
        data_types = previous_query.data_types
        if len(data_types) > 1:
            raise ValueError('SignificantDE can only be used with one data type')
        data_type = data_types[0]

        # get the DE result for the case vs control comparison
        de_res = pipeline.get_de_results(data_type, self.case, self.control)

        # filter results
        pval_col = 'padj_%s_vs_%s' % (self.case, self.control)
        fc_col = 'FC_%s_vs_%s' % (self.case, self.control)
        if self.fc_lte is None and self.fc_gte is None:
            # filter by p-value only
            filtered = de_res[de_res[pval_col] <= self.pval]
        else:
            # filter by p-value and fold-change
            fc_lte = -np.inf if self.fc_lte is None else self.fc_lte
            fc_gte = np.inf if self.fc_gte is None else self.fc_gte
            filtered = de_res[
                (de_res[pval_col] <= self.pval) & ((de_res[fc_col] <= fc_lte) | (de_res[fc_col] >= fc_gte))]

        # merge with previous selection result
        previous_res = previous_query.get_result()
        merged = pd.merge(previous_res, filtered, how='inner', left_index=True, right_index=True)

        # if the sort order is not specified, take N/2 from the top and bottom halves of FC result, and combine them
        if self.ascending is None:
            N = int(self.N / 2)
            top = merged.sort_values(fc_col, ascending=False).head(N)
            bottom = merged.sort_values(fc_col, ascending=True).head(N)
            combined = pd.concat([top, bottom], axis=0)
            final_res = combined.sort_values(fc_col, ascending=False)
        else:
            # else take the top N result according to the sort order
            final_res = merged.sort_values(fc_col, ascending=self.ascending).head(self.N)
        self.result = final_res


class Info(Query):
    def run(self, pipeline, previous_query):
        previous_res = previous_query.get_result()
        copy_df = previous_res.copy()
        node_info = []
        node_images = []
        node_links = []
        for node_id, node_data in copy_df.iterrows():
            data_type = node_data[QUERY_DATA_TYPE]
            res = get_info(node_id, data_type)
            node_info.append(res['infos'])
            node_images.append(res['images'])
            node_links.append(res['links'])

        copy_df['infos'] = node_info
        copy_df['images'] = node_images
        copy_df['links'] = node_links
        self.result = copy_df



from loguru import logger

from .constants import GENES, PROTEINS, COMPOUNDS, REACTIONS, PATHWAYS, SMALL, GENOMICS, INFERENCE_T_TEST, \
    INFERENCE_DESEQ, INFERENCE_LIMMA
from .pipelines import WebOmicsInference


class AnalysisPipeline(object):
    def __init__(self, mapping):
        self.mapping = mapping
        self.analyses = {
            GENES: [],
            PROTEINS: [],
            COMPOUNDS: [],
            REACTIONS: [],
            PATHWAYS: []
        }

    def run_de(self, selected_method, data_type, case, control):
        method_to_run = {
            INFERENCE_T_TEST: TTestAnalysis,
            INFERENCE_DESEQ: DESeq2Analysis,
            INFERENCE_LIMMA: LimmaAnalysis
        }
        M = method_to_run[selected_method]

        # check if already exists
        found = False
        for analysis in self.analyses[data_type]:
            if analysis.case == case and analysis.control == control:
                found = True

        if not found:
            data_df, design_df = self.mapping.get_dfs(data_type)
            analysis = M(data_df, design_df, data_type, case, control)
            analysis.run()
            self.analyses[data_type].append(analysis)
        else:
            logger.warning('DE analysis for data_type=%s case=%s control=%s already exists' % (
                data_type, case, control
            ))

    def get_de_analysis(self, data_type, case, control, method=None):
        if method is not None:
            assert method in [INFERENCE_T_TEST, INFERENCE_DESEQ, INFERENCE_LIMMA]

        found = None
        for analysis in self.analyses[data_type]:
            # select only analysis for a particular case and control, if specified
            if analysis.case != case or analysis.control != control:
                continue

            # select method, if specified
            if method is not None and analysis.get_analysis_type() != method:
                continue

            # return the first result we found
            found = analysis
        return found

    def get_de_results(self, data_type, case, control, method=None):
        # get results dataframe for the DE analysis
        analysis = self.get_de_analysis(data_type, case, control, method=method)
        res = analysis.get_results()

        # rename the column 'log2FoldChange' to 'FC'
        res = res.rename(columns={
            'log2FoldChange': 'FC'
        })

        # add suffix to 'padj' and 'FC' columns
        suffix = '_%s_vs_%s' % (analysis.case, analysis.control)
        res = res.add_suffix(suffix)
        return res.sort_index()


class CaseControlAnalysis():
    def __init__(self, data_df, design_df, data_type, case, control):
        self.data_df = data_df
        self.design_df = design_df
        self.data_type = data_type
        self.case = case
        self.control = control
        self.results = None
        self.wi = None

    def __repr__(self):
        return 'CaseControlAnalysis %s: %s_vs_%s' % (self.data_type, self.case, self.control)

    def get_results(self):
        return self.results

    def run(self):
        raise NotImplementedError

    def get_analysis_type(self):
        raise NotImplementedError


class TTestAnalysis(CaseControlAnalysis):
    def __init__(self, data_df, design_df, data_type, case, control):
        super().__init__(data_df, design_df, data_type, case, control)

    def run(self):
        min_replace = SMALL
        self.wi = WebOmicsInference(self.data_df, self.design_df, self.data_type, min_value=min_replace)
        self.results = self.wi.run_ttest(self.case, self.control)

    def get_analysis_type(self):
        return INFERENCE_T_TEST

    def __repr__(self):
        return 't-test %s: %s_vs_%s' % (self.data_type, self.case, self.control)


class DESeq2Analysis(CaseControlAnalysis):
    def __init__(self, data_df, design_df, data_type, case, control):
        super().__init__(data_df, design_df, data_type, case, control)

    def run(self):
        assert self.data_type == GENOMICS
        MIN_VALUE = 1
        KEEP_THRESHOLD = 10
        REPLACE_MEAN = False
        self.wi = WebOmicsInference(self.data_df, self.design_df, self.data_type, min_value=MIN_VALUE,
                                    replace_mean=REPLACE_MEAN)
        try:
            pd_df, rld_df, res_ordered = self.wi.run_deseq(KEEP_THRESHOLD, self.case, self.control)
            self.results = pd_df[['padj', 'log2FoldChange']]
        except Exception as e:
            logger.warning('Failed to run DESeq2: %s' % str(e))

    def get_analysis_type(self):
        return INFERENCE_DESEQ

    def __repr__(self):
        return 'DESeq2 %s: %s_vs_%s' % (self.data_type, self.case, self.control)


class LimmaAnalysis(CaseControlAnalysis):
    def __init__(self, data_df, design_df, data_type, case, control):
        super().__init__(data_df, design_df, data_type, case, control)

    def run(self):
        min_replace = SMALL
        try:
            self.wi = WebOmicsInference(self.data_df, self.design_df, self.data_type, min_value=min_replace)
            self.results = self.wi.run_limma(self.case, self.control)
        except Exception as e:
            logger.warning('Failed to run limma: %s' % str(e))

    def get_analysis_type(self):
        return INFERENCE_LIMMA

    def __repr__(self):
        return 'limma %s: %s_vs_%s' % (self.data_type, self.case, self.control)



import gzip
import logging
import math
import os
import pathlib
import pickle
import sys
import zipfile

from loguru import logger
import requests
from tqdm import tqdm

from pyMultiOmics.constants import TRUNCATE_LIMIT


def create_if_not_exist(out_dir):
    if not os.path.exists(out_dir) and len(out_dir) > 0:
        logger.info('Created %s' % out_dir)
        pathlib.Path(out_dir).mkdir(parents=True, exist_ok=True)


def save_obj(obj, filename):
    """
    Save object to file
    :param obj: the object to save
    :param filename: the output file
    :return: None
    """
    out_dir = os.path.dirname(filename)
    create_if_not_exist(out_dir)
    logger.info('Saving %s to %s' % (type(obj), filename))
    with gzip.GzipFile(filename, 'w') as f:
        pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_obj(filename):
    """
    Load saved object from file
    :param filename: The file to load
    :return: the loaded object
    """
    try:
        with gzip.GzipFile(filename, 'rb') as f:
            return pickle.load(f)
    except OSError:
        logging.getLogger().warning('Old, invalid or missing pickle in %s. Please regenerate this file.' % filename)
        return None


def download_file(url, out_file=None):
    r = requests.get(url, stream=True)
    total_size = int(r.headers.get('content-length', 0));
    block_size = 1024
    current_size = 0

    if out_file is None:
        out_file = url.rsplit('/', 1)[-1]  # get the last part in url
    logger.info('Downloading %s' % out_file)

    with open(out_file, 'wb') as f:
        for data in tqdm(r.iter_content(block_size), total=math.ceil(total_size // block_size), unit='KB',
                         unit_scale=True):
            current_size += len(data)
            f.write(data)
    assert current_size == total_size
    return out_file


def extract_zip_file(in_file, delete=True):
    logger.info('Extracting %s' % in_file)
    with zipfile.ZipFile(file=in_file) as zip_file:
        for file in tqdm(iterable=zip_file.namelist(), total=len(zip_file.namelist())):
            zip_file.extract(member=file)

    if delete:
        logger.info('Deleting %s' % in_file)
        os.remove(in_file)


def as_list(value):
    if isinstance(value, str): # if string turn this into a list item
        value = [value]
    else: # other types, check if  they're iterable
        try:
            iter(value)
        except TypeError: # not iterable
            value = [value]
        else: # iterable
            value = list(value)
    return value


def set_log_level(level, remove_id=None):
    if remove_id is None:
        try:
            logger.remove(0)  # try to remove the default handler with id 0
        except ValueError:  # no default handler has been set
            pass
    else:
        logger.remove(remove_id)  # remove previously set handler by id

    # add new handler at the desired log level
    new_handler_id = logger.add(sys.stderr, level=level)
    return new_handler_id


def set_log_level_warning(remove_id=None):
    return set_log_level(logging.WARNING, remove_id=remove_id)


def set_log_level_info(remove_id=None):
    return set_log_level(logging.INFO, remove_id=remove_id)


def set_log_level_debug(remove_id=None):
    return set_log_level(logging.DEBUG, remove_id=remove_id)


def add_log_file(log_path, level):
    logger.add(log_path, level=level)


def truncate(my_str):
    my_str = (my_str[:TRUNCATE_LIMIT] + '...') if len(my_str) > TRUNCATE_LIMIT else my_str
    return my_str



import json

import networkx as nx
import pandas as pd
from loguru import logger

from .common import as_list
from .constants import REACTIONS, PROTEOMICS, METABOLOMICS, GENOMICS, TRANSCRIPTOMICS, PATHWAYS, DataTypeDict, \
    COMPOUND_DATABASE_CHEBI, MAPPING, PKS, IDS, NA, GENES_TO_PROTEINS, PROTEINS_TO_REACTIONS, COMPOUNDS_TO_REACTIONS, \
    REACTIONS_TO_PATHWAYS
from .functions import reactome_mapping
from .query import QueryBuilder, Connected


class Mapper():
    def __init__(self, species, metabolic_pathway_only=True,
                 compound_database_str=COMPOUND_DATABASE_CHEBI,
                 include_related_chebi=False):
        self.species_list = [species]
        self.metabolic_pathway_only = metabolic_pathway_only
        self.compound_database_str = compound_database_str
        self.include_related_chebi = include_related_chebi

        self.G = None

        self.gene_df = None
        self.gene_design = None

        self.protein_df = None
        self.protein_design = None

        self.compound_df = None
        self.compound_design = None

    def set_gene(self, gene_df, gene_design):
        self.gene_df = gene_df
        self.gene_design = gene_design
        return self

    def set_protein(self, protein_df, protein_design):
        self.protein_df = protein_df
        self.protein_design = protein_design
        return self

    def set_compound(self, compound_df, compound_design):
        self.compound_df = compound_df
        self.compound_design = compound_design
        return self

    def get_dfs(self, data_type):
        data_df = None
        design_df = None
        if data_type == GENOMICS:
            data_df = self.gene_df
            design_df = self.gene_design
        elif data_type == PROTEOMICS:
            data_df = self.protein_df
            design_df = self.protein_design
        elif data_type == METABOLOMICS:
            data_df = self.compound_df
            design_df = self.compound_design
        return data_df, design_df

    def build(self):

        # map different omics entities to reactome
        results = reactome_mapping(self.gene_df, self.protein_df, self.compound_df, self.compound_database_str,
                                   self.species_list, self.metabolic_pathway_only, self.include_related_chebi)
        designs = {
            GENOMICS: self.gene_design,
            PROTEOMICS: self.protein_design,
            METABOLOMICS: self.compound_design
        }

        # create network graphs, can be retrieved using self._get_graph()
        self.G = nx.Graph()

        # add nodes and edges from mapping results to the graph
        self._add_nodes(results, designs)
        self._add_edges(results)
        self._cleanup()

        logger.info('Created a multi-omics network with %d nodes and %d edges' %
                    (len(self.G.nodes), len(self.G.edges)))
        logger.info('node_counts = %s' % self.num_nodes())
        return self

    def get_node(self, node_id):
        graph = self._get_graph()
        return graph.nodes[node_id]

    def get_node_type(self, node_id):
        node = self.get_node(node_id)
        return node['data_type']

    def get_display_name(self, node_id):
        node = self.get_node(node_id)
        return node['display_name']

    def is_observed(self, node_id):
        node = self.get_node(node_id)
        return node['observed']

    def get_nodes(self, types=None):
        graph = self._get_graph()
        nodes = list(graph.nodes(data=True))
        if types is None:
            return nodes
        types = as_list(types)

        # need to filter
        # below, node[0] is the id, while node[1] is the data dict
        results = filter(lambda node: node[1]['data_type'] in types, nodes)
        return list(results)

    def num_nodes(self, types=None):
        if types is None:
            types = [GENOMICS, TRANSCRIPTOMICS, PROTEOMICS, METABOLOMICS, REACTIONS, PATHWAYS]
        else:
            types = as_list(types)
        results = {DataTypeDict[t]: len(self.get_nodes(types=t)) for t in types}
        return results

    def get_neighbours(self, query_id, include_types=None, exclude_types=None):
        graph = self._get_graph()
        results = list(graph.neighbors(query_id))

        # filter for types to be included
        if include_types is not None:
            include_types = as_list(include_types)
            results = filter(lambda node: self.get_node_type(node) in include_types, results)

        # filter for types to be excluded
        if exclude_types is not None:
            exclude_types = as_list(exclude_types)
            results = filter(lambda node: self.get_node_type(node) not in exclude_types, results)

        return list(results)

    def get_connected(self, query_id, dest_type=None, observed=None):
        """
        Retrieve mapped entities that are connected to query_id
        :param query_id: the node ID to query
        :param dest_type: connected entity types, e.g. GENES, PROTEINS, COMPOUNDS, REACTIONS, PATHWAYS
        :param observed: True to get entities observed in the data, False to retrieve everything mapped by Reactome
        :return: a dataframe of connected entities
        """
        res = QueryBuilder(self) \
            .add(SingleEntity(query_id)) \
            .add(Connected(data_type=dest_type, observed=observed)) \
            .run()
        return res

    def _get_graph(self):
        return self.G

    def _json_str_to_df(self, json_str):
        res = json.loads(json_str)
        df = pd.DataFrame(res)
        return df

    def _filter_series(self, s, includes):
        if len(includes) == 0:
            return None

        new_dict = {}
        for index, value in s.items():
            if index in includes:
                new_dict[index] = value
        new_s = pd.Series(new_dict)
        return new_s

    def _insert_edge_if_node_exists(self, source_id, dest_id):
        graph = self._get_graph()
        if source_id in graph.nodes and dest_id in graph.nodes:
            graph.add_edge(source_id, dest_id)

    def _add_nodes(self, results, designs):
        graph = self._get_graph()
        data_types = [GENOMICS, PROTEOMICS, METABOLOMICS, REACTIONS, PATHWAYS]
        for data_type in data_types:

            display_data_type = MAPPING[data_type]
            logger.info('Processing nodes: %s' % display_data_type)

            # get identifier and display name columns
            id_col = PKS[data_type]
            name_col = IDS[data_type]

            # convert the dataframes that we need
            data_df = self._json_str_to_df(results[data_type])
            try:
                design_df = designs[data_type]
                sample_names = design_df.index.values.tolist()
            except KeyError:
                sample_names = []
            except AttributeError:
                sample_names = []

            # insert each row in dataframe into a network graph G
            for idx, row in data_df.iterrows():

                # get node identifier
                identifier = row[id_col]
                if identifier == NA:
                    continue

                # select only sample names columns to get the measurements
                observed = row['obs']
                measurements = None
                if observed:
                    measurements = self._filter_series(row, sample_names)

                # create node info dictionary
                node_info = {
                    'observed': observed,
                    'display_name': row[name_col],
                    'data_type': data_type,
                    'display_data_type': display_data_type,
                    'measurements': measurements,
                }

                # insert node into two separate graphs: G and G_complete
                # G_complete contains both actual and inferred nodes
                # G contains only the actual nodes
                graph.add_node(identifier, **node_info)

    def _add_edges(self, results):
        data_types = [GENES_TO_PROTEINS, PROTEINS_TO_REACTIONS, COMPOUNDS_TO_REACTIONS, REACTIONS_TO_PATHWAYS]
        for data_type in data_types:
            display_data_type = MAPPING[data_type]
            logger.info('Processing edges: %s' % display_data_type)

            df = self._json_str_to_df(results[data_type])
            for idx, row in df.iterrows():
                source_id, dest_id = row.values.tolist()

                # only add edges between nodes that already exist (have been added before)
                self._insert_edge_if_node_exists(source_id, dest_id)

    def _cleanup(self):

        # get all reactions nodes from the graph
        reactions = self.get_nodes(types=REACTIONS)
        to_delete = []
        for reaction_id, reaction_data in reactions:
            # FIXME: we shouldn't need to do this?
            # delete a reaction where the display_name is the same as the id
            # this would be the case when this reaction isn't found in the query to map reaction -> pathways
            # i.e. when only metabolic pathways are selected
            # in this case, this reaction actually shouldn't be included in the JSON results, but they are.
            # as a workaround we'd delete them here
            reaction_name = reaction_data['display_name']
            if reaction_id == reaction_name:
                to_delete.append(reaction_id)

        logger.debug('Deleting %d reactions that should not be there' % len(to_delete))
        logger.debug(to_delete[0:100])
        graph = self._get_graph()
        graph.remove_nodes_from(to_delete)



import numpy as np
import pandas as pd
from loguru import logger
from scipy import stats
from sklearn import preprocessing
from sklearn.decomposition import PCA
from statsmodels.sandbox.stats.multicomp import multipletests

from .constants import GROUP_COL, IDS, PKS


class WebOmicsInference(object):
    def __init__(self, data_df, design_df, data_type, remove_cols=None, min_value=0, replace_mean=True):

        data_df = data_df.copy()

        # remove all the default columns from dataframe if nothing provided
        if remove_cols is None:
            remove_cols = ['padj_', 'FC_', 'significant_', 'obs', PKS[data_type], IDS[data_type]]
        remove_cols = tuple([x.lower() for x in remove_cols])
        to_drop = list(filter(lambda x: x.lower().startswith(remove_cols), data_df.columns))
        df = data_df.drop(to_drop, axis=1)

        # remove rows that are all NAs and all 0s
        df = df.dropna(how='all')
        df = df.loc[~(df == 0).all(axis=1)]

        self.data_df = df
        self.design_df = design_df
        self.data_type = data_type

        # data imputation:
        # - if all zero in group then replace with min_value
        # - replace any other zeros with mean of group
        self._impute_data(min_value, replace_mean=replace_mean)

    # def heatmap(self, N=None, standardize=True, log=False):
    #     if standardize:
    #         data_df = self.standardize_df(self.data_df, log=log)
    #     else:
    #         data_df = self.data_df
    #     if N is not None:
    #         plt.matshow(data_df[0:N])
    #     else:
    #         plt.matshow(data_df)
    #     plt.colorbar()

    def standardize_df(self, data_df, log=False, axis=1):
        if data_df.empty:
            return data_df
        data_df = data_df.copy()

        # log only if all the values are positive
        # assume if there's a negative value, the data has been pre-processed, so don't do it again
        data_arr = np.array(data_df)
        if np.all(data_arr >= 0) and log:
            data_arr = np.log(data_arr)

        # center data to have 0 mean and unit variance for heatmap and pca
        scaled_data = preprocessing.scale(data_arr, axis=axis)

        # set the values back to the dataframe
        sample_names = data_df.columns
        data_df[sample_names] = scaled_data
        return data_df

    def run_deseq(self, keep_threshold, case, control):
        logger.info('DeSEQ2 support is not available, please install rpy2 extra package')

    def run_ttest(self, case, control):
        logger.info('t-test case is %s, control is %s' % (case, control))
        count_data = self.data_df
        col_data = self.design_df
        sample_group = col_data[col_data[GROUP_COL] == case]
        case_data = count_data[sample_group.index]
        sample_group = col_data[col_data[GROUP_COL] == control]
        control_data = count_data[sample_group.index]

        nrow, _ = count_data.shape
        pvalues = []
        lfcs = []
        indices = []

        # if there's a negative number, assume the data has been logged, so don't do it again
        log = np.all(count_data.values >= 0)

        # T-test for the means of two independent samples
        for i in range(nrow):

            case = case_data.iloc[i, :].values
            control = control_data.iloc[i, :].values
            idx = count_data.index[i]

            # remove 0 values, which were originally NA when exported from PiMP
            case = case[case != 0]
            control = control[control != 0]

            # log the data if it isn't already logged
            if log:
                case_log = np.log2(case)
                control_log = np.log2(control)
            else:
                case_log = case
                control_log = control

            statistics, pvalue = stats.ttest_ind(case_log, control_log)
            if not np.isnan(pvalue):
                lfc = np.mean(case_log) - np.mean(control_log)
                pvalues.append(pvalue)
                lfcs.append(lfc)
                indices.append(idx)

        # correct p-values
        reject, pvals_corrected, _, _ = multipletests(pvalues, method='fdr_bh')
        result_df = pd.DataFrame({
            'padj': pvals_corrected,
            'log2FoldChange': lfcs
        }, index=indices)
        return result_df

    def run_limma(self, case, control):
        logger.info('limma support is not available, please install rpy2 extra package')

    def get_pca(self, rld_df, n_components, plot=False):
        df = rld_df.transpose()
        pca = PCA(n_components=n_components)
        X = pca.fit_transform(df)

        # if plot:
        #     fig, ax = plt.subplots()
        #     ax.scatter(X[:, 0], X[:, 1])
        #     for i, txt in enumerate(df.index):
        #         ax.annotate(txt, (X[i, 0], X[i, 1]))
        #     plt.tight_layout()
        #     fn = '{uuid}.png'.format(uuid=uuid.uuid4())
        #     plt.save(fn)

        cumsum = np.cumsum(pca.explained_variance_ratio_)
        return X, cumsum

    def _impute_data(self, min_value, replace_mean=True):
        if self.design_df is not None:
            grouping = self.design_df.groupby('group')
            for group, samples in grouping.groups.items():

                # If all zero in group then replace with minimum
                temp = self.data_df.loc[:, samples]
                temp = (temp == 0).all(axis=1)
                self.data_df.loc[temp, samples] = min_value

                if replace_mean:
                    # replace any other zeros with mean of group
                    subset_df = self.data_df.loc[:, samples]
                    self.data_df.loc[:, samples] = subset_df.mask(subset_df == 0, subset_df.mean(axis=1), axis=0)
                else:
                    # replace all 0s with min_value
                    self.data_df = self.data_df.replace(0, min_value)



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()
setup(
    name='pyMultiOmics',
    version='1.0.0',
    author='Joe Wandy',
    author_email='joe.wandy@glasgow.ac.uk',
    description='A Python package for multi-omics data integration and analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/glasgowcompbio/pyMultiOmics',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    packages=find_packages(),
    package_data={
        'pyMultiOmics': [
            'data/*.*',
        ],
    },
    install_requires=['numpy', 'scipy', 'pandas', 'seaborn', 'scikit-learn', 'matplotlib', 'plotly',
                      'statsmodels', 'loguru', 'requests', 'neo4j-driver==1.7.6',
                      'tqdm', 'pillow', 'jupyterlab', 'bioservices', 'networkx', 'tzlocal', 'pals-pathway'],
)



import os


def get_data_path(data_dir, file_name):
    return os.path.abspath(os.path.join(data_dir, '%s' % file_name))


BASE_DIR = os.path.dirname(os.path.dirname(__file__))
DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')
EXTERNAL_COMPOUND_NAMES = get_data_path(DATA_DIR, 'compound_names.p')
EXTERNAL_KEGG_TO_CHEBI = get_data_path(DATA_DIR, 'kegg_to_chebi.p')
EXTERNAL_GENE_NAMES = get_data_path(DATA_DIR, 'gene_names.p')
EXTERNAL_GO_DATA = get_data_path(DATA_DIR, 'go_data.p')

CHEBI_BFS_RELATION_DICT = get_data_path(DATA_DIR, 'chebi_bfs_relation_dict.pkl')
CHEBI_RELATION_TSV = get_data_path(DATA_DIR, 'relation.tsv')

ARABIDOPSIS_THALIANA = 'Arabidopsis thaliana'
BOS_TAURUS = 'Bos taurus'
CAENORHABDITIS_ELEGANS = 'Caenorhabditis elegans'
CANIS_LUPUS_FAMILIARIS = 'Canis lupus familiaris'
DANIO_RERIO = 'Danio rerio'
DICTYOSTELIUM_DISCOIDEUM = 'Dictyostelium discoideum'
DROSOPHILA_MELANOGASTER = 'Drosophila melanogaster'
GALLUS_GALLUS = 'Gallus gallus'
HOMO_SAPIENS = 'Homo sapiens'
MUS_MUSCULUS = 'Mus musculus'
ORYZA_SATIVA = 'Oryza sativa'
RATTUS_NORVEGICUS = 'Rattus norvegicus'
SACCHAROMYCES_CEREVISIAE = 'Saccharomyces cerevisiae'
SUS_SCROFA = 'Sus scrofa'

NA = '-'
ALL = 'ALL'

GENE_PK = 'gene_pk'
PROTEIN_PK = 'protein_pk'
COMPOUND_PK = 'compound_pk'
REACTION_PK = 'reaction_pk'
PATHWAY_PK = 'pathway_pk'

GENE_ID = 'gene_id'
PROTEIN_ID = 'protein_id'
COMPOUND_ID = 'compound_id'
REACTION_ID = 'reaction_id'
PATHWAY_ID = 'pathway_id'

GENOMICS = 0
TRANSCRIPTOMICS = 1
PROTEOMICS = 2
METABOLOMICS = 3
REACTIONS = 4
PATHWAYS = 5
GENES_TO_PROTEINS = 6
PROTEINS_TO_REACTIONS = 7
COMPOUNDS_TO_REACTIONS = 8
REACTIONS_TO_PATHWAYS = 9
MULTI_OMICS = 10

# aliases
GENES = GENOMICS
TRANSCRIPTS = TRANSCRIPTOMICS
PROTEINS = PROTEOMICS
COMPOUNDS = METABOLOMICS

DataType = (
    (GENOMICS, 'Genomics'),
    (TRANSCRIPTOMICS, 'Transcriptomics'),
    (PROTEOMICS, 'Proteomics'),
    (METABOLOMICS, 'Metabolomics'),
    (REACTIONS, 'Reactions'),
    (PATHWAYS, 'Pathways')
)
DataTypeDict = dict(DataType)

MAPPING = {
    GENOMICS: 'genes',
    PROTEOMICS: 'proteins',
    METABOLOMICS: 'compounds',
    REACTIONS: 'reactions',
    PATHWAYS: 'pathways',
    GENES_TO_PROTEINS: 'gene_proteins',
    PROTEINS_TO_REACTIONS: 'protein_reactions',
    COMPOUNDS_TO_REACTIONS: 'compound_reactions',
    REACTIONS_TO_PATHWAYS: 'reaction_pathways'
}

PKS = {
    GENOMICS: GENE_PK,
    PROTEOMICS: PROTEIN_PK,
    METABOLOMICS: COMPOUND_PK,
    REACTIONS: REACTION_PK,
    PATHWAYS: PATHWAY_PK
}

IDS = {
    GENOMICS: GENE_ID,
    PROTEOMICS: PROTEIN_ID,
    METABOLOMICS: COMPOUND_ID,
    REACTIONS: REACTION_ID,
    PATHWAYS: PATHWAY_ID
}

COMPOUND_DATABASE_KEGG = 'KEGG'
COMPOUND_DATABASE_CHEBI = 'ChEBI'

IDENTIFIER_COL = 'Identifier'
PADJ_COL_PREFIX = 'padj_'
FC_COL_PREFIX = 'FC_'
SAMPLE_COL = 'sample'
GROUP_COL = 'group'
FACTOR_COL = 'factor'
DEFAULT_GROUP_NAME = 'default'

PIMP_PEAK_ID_COL = 'Peak id'
SMALL = 1E-6

# Constants used in the Inference page

INFERENCE_LOADED = 0
INFERENCE_T_TEST = 1
INFERENCE_CORRELATION = 2
INFERENCE_PCA = 3
INFERENCE_PALS = 4
INFERENCE_ORA = 5
INFERENCE_GSEA = 6
INFERENCE_DESEQ = 7
INFERENCE_LIMMA = 8
INFERENCE_REACTOME = 9

T_TEST_THRESHOLD = 0.05

QUERY_ENTITY_ID = 'entity_id'
QUERY_OBSERVED = 'observed'
QUERY_DATA_TYPE = 'data_type'
QUERY_NODE_ID = 'node_id'
QUERY_DISPLAY_NAME = 'display_name'
QUERY_SOURCE_ID = 'source_id'

TRUNCATE_LIMIT = 10000